"""This module contains the resolvers for user settings."""

from sqlmodel import Session, select

from lightly_purple.server.models import Setting, SettingView


class SettingsResolver:
    """Resolver for user settings."""

    def __init__(self, session: Session):
        """Initialize the resolver with a database session.

        Args:
            session: The database session.
        """
        self.session = session

    def get_settings(self) -> SettingView:
        """Get current settings.

        Returns:
            The current settings.
        """
        statement = select(Setting)
        result = self.session.exec(statement).first()

        # If no settings exist, create default settings
        if result is None:
            result = Setting()
            self.session.add(result)
            self.session.commit()
            self.session.refresh(result)

        return SettingView(
            setting_id=result.setting_id,
            grid_view_sample_rendering=result.grid_view_sample_rendering,
        )

    def set_settings(self, settings: SettingView) -> SettingView:
        """Update settings.

        Args:
            settings: New settings to apply.

        Returns:
            Updated settings.
        """
        current_settings = self.session.exec(select(Setting)).first()
        if current_settings is None:
            current_settings = Setting()
            self.session.add(current_settings)

        # Update grid view sample rendering
        if settings.grid_view_sample_rendering is not None:
            current_settings.grid_view_sample_rendering = (
                settings.grid_view_sample_rendering
            )

        self.session.commit()
        self.session.refresh(current_settings)

        return SettingView(
            setting_id=current_settings.setting_id,
            grid_view_sample_rendering=current_settings.grid_view_sample_rendering,
        )



"""Embedding manager for dataset processing."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from uuid import UUID

from sqlmodel import Session

from lightly_purple.dataset.embedding_generator import EmbeddingGenerator
from lightly_purple.server.models import (
    EmbeddingModel,
)
from lightly_purple.server.models.sample_embedding import SampleEmbeddingInput
from lightly_purple.server.resolvers.embedding_model import (
    EmbeddingModelResolver,
)
from lightly_purple.server.resolvers.sample import SampleResolver
from lightly_purple.server.resolvers.sample_embedding import (
    SampleEmbeddingResolver,
)


class EmbeddingManagerProvider:
    """Provider for the EmbeddingManager singleton instance."""

    _instance: EmbeddingManager | None = None

    @classmethod
    def get_embedding_manager(cls, session: Session | None) -> EmbeddingManager:
        """Get the singleton instance of EmbeddingManager.

        Args:
            session: A valid database session. Required if the instance is not
                     yet created.

        Returns:
            The singleton instance of EmbeddingManager.

        Raises:
            ValueError: If no instance exists and no session is provided.
        """
        if cls._instance is None:
            if session is None:
                raise ValueError(
                    "A valid session is required to create an EmbeddingManager."
                )
            cls._instance = EmbeddingManager(session)
        return cls._instance


@dataclass
class TextEmbedQuery:
    """Parameters for text embedding generation."""

    text: str
    embedding_model_id: UUID | None = None


class EmbeddingManager:
    """Manages embedding models and handles embedding generation and storage."""

    def __init__(
        self,
        session: Session,
    ) -> None:
        """Initialize the embedding manager.

        Args:
            session: Database session for resolver operations.
        """
        self._embedding_model_resolver = EmbeddingModelResolver(session)
        self._sample_embedding_resolver = SampleEmbeddingResolver(session)
        self._sample_resolver = SampleResolver(session)
        self._models: dict[UUID, EmbeddingGenerator] = {}
        self._default_model_id: UUID | None = None

    def register_embedding_model(
        self,
        dataset_id: UUID,
        embedding_generator: EmbeddingGenerator,
        set_as_default: bool = False,
    ) -> EmbeddingModel:
        """Register an embedding model in the database.

        The model is stored in an internal dictionary for later use.
        The model is set as default if requested or if it's the first model.

        Args:
            dataset_id: The ID of the dataset to associate with the model.
            embedding_generator: The model implementation used for embeddings.
            set_as_default: Whether to set this model as the default.

        Returns:
            The created EmbeddingModel.
        """
        # Create embedding model record in the database.
        created_model = self._embedding_model_resolver.create(
            embedding_generator.get_embedding_model_input(dataset_id=dataset_id)
        )
        model_id = created_model.embedding_model_id

        # Store the model in our dictionary
        self._models[model_id] = embedding_generator

        # Set as default if requested or if it's the first model
        if set_as_default or self._default_model_id is None:
            self._default_model_id = model_id

        return created_model

    def embed_text(self, text_query: TextEmbedQuery) -> list[float]:
        """Generate an embedding for a text sample.

        Args:
            text_query: Text embedding query containing text and model ID.

        Returns:
            A list of floats representing the generated embedding.
        """
        model_id = text_query.embedding_model_id or self._default_model_id
        if model_id is None:
            raise ValueError(
                "No embedding model specified and no default model set."
            )

        model = self._models.get(model_id)
        if model is None:
            raise ValueError(f"Embedding model with ID {model_id} not found.")

        return model.embed_text(text_query.text)

    def embed_images(
        self,
        sample_ids: list[UUID],
        embedding_model_id: UUID | None = None,
    ) -> None:
        """Generate and store embeddings for samples.

        Args:
            sample_ids: List of sample IDs to generate embeddings for.
            embedding_model_id: ID of the model to use. Uses default if None.

        Raises:
            ValueError: If no embedding model is registered or provided model
            ID doesn't exist.
        """
        model_id = embedding_model_id or self._default_model_id
        if not model_id:
            raise ValueError("No default embedding model registered.")

        if model_id not in self._models:
            raise ValueError(f"No embedding model found with ID {model_id}")

        # Query image filenames from the database.
        sample_id_to_filepath = {
            sample.sample_id: Path(sample.file_path_abs)
            for sample in self._sample_resolver.get_many_by_id(
                sample_ids=sample_ids
            )
        }

        # Extract filepaths in the same order as sample_ids.
        filepaths = [
            sample_id_to_filepath[sample_id] for sample_id in sample_ids
        ]

        # Generate embeddings for the samples.
        embeddings = self._models[model_id].embed_images(filepaths=filepaths)

        # Convert to SampleEmbeddingInput objects.
        sample_embeddings = [
            SampleEmbeddingInput(
                sample_id=sample_id,
                embedding_model_id=model_id,
                embedding=embedding,
            )
            for sample_id, embedding in zip(sample_ids, embeddings)
        ]

        # Store the embeddings in the database.
        self._sample_embedding_resolver.create_many(sample_embeddings)



"""Handler for database operations related to sample embeddings."""

from __future__ import annotations

from sqlmodel import Session

from lightly_purple.server.models import SampleEmbedding
from lightly_purple.server.models.sample_embedding import SampleEmbeddingInput


class SampleEmbeddingResolver:
    """Resolver for the sample embedding model."""

    def __init__(self, session: Session):
        """Initializes the SampleEmbeddingResolver."""
        self.session = session

    def create(self, sample_embedding: SampleEmbeddingInput) -> SampleEmbedding:
        """Create a new SampleEmbedding in the database."""
        db_sample_embedding = SampleEmbedding.model_validate(sample_embedding)
        self.session.add(db_sample_embedding)
        self.session.commit()
        self.session.refresh(db_sample_embedding)
        return db_sample_embedding

    def create_many(
        self, sample_embeddings: list[SampleEmbeddingInput]
    ) -> None:
        """Create many sample embeddings in a single database commit."""
        db_sample_embeddings = [
            SampleEmbedding.model_validate(e) for e in sample_embeddings
        ]
        self.session.bulk_save_objects(db_sample_embeddings)
        self.session.commit()



"""This module defines the AnnotationTask model."""

from datetime import datetime
from enum import Enum
from uuid import UUID, uuid4

from sqlmodel import Field, SQLModel


class AnnotationType(str, Enum):
    """The type of annotation task."""

    BBOX = "bbox"
    CLASSIFICATION = "classification"
    SEMANTIC_SEGMENTATION = "semantic_segmentation"
    INSTANCE_SEGMENTATION = "instance_segmentation"


class AnnotationTask(SQLModel, table=True):
    """The annotation task model."""

    __tablename__ = "annotation_tasks"
    annotation_task_id: UUID = Field(default_factory=uuid4, primary_key=True)
    name: str
    created_at: datetime = Field(default_factory=datetime.utcnow)
    annotation_type: AnnotationType
    is_prediction: bool



"""This module contains the Tag model and related enumerations."""

from typing import Literal, Optional
from uuid import UUID

from sqlmodel import SQLModel

# TagKind is the kind of tag we support.
TagKind = Literal["sample", "annotation"]


class TagBase(SQLModel):
    """Base class for the Tag model."""

    name: str
    description: Optional[str] = ""
    kind: TagKind = "sample"


class TagInput(TagBase):
    """Tag model when creating."""

    dataset_id: UUID


class TagInputBody(TagBase):
    """Tag model when creating."""


class TagUpdate(TagBase):
    """Tag model when updating."""


class TagUpdateBody(TagBase):
    """Tag model when updating."""

    dataset_id: Optional[UUID] = None  # type: ignore


class TagView(TagBase):
    """Tag model when retrieving."""

    tag_id: UUID
    kind: TagKind



"""Initialize environment variables for the dataset module."""

import os

from dotenv import load_dotenv

load_dotenv()

PURPLE_PROTOCOL: str = os.getenv("PURPLE_PROTOCOL", "http")
PURPLE_PORT: int = int(os.getenv("PURPLE_PORT", "8001"))
PURPLE_HOST: str = os.getenv("PURPLE_HOST", "localhost")
PURPLE_DEBUG: str = os.getenv("PURPLE_DEBUG", "FALSE")

APP_URL = f"{PURPLE_PROTOCOL}://{PURPLE_HOST}:{PURPLE_PORT}"



"""This module contains the API routes for managing tags."""

from __future__ import annotations

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Path, Query
from pydantic import BaseModel
from sqlalchemy.exc import IntegrityError
from sqlmodel import Field, Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import Dataset, Tag
from lightly_purple.server.models.tag import (
    TagInput,
    TagInputBody,
    TagUpdate,
    TagUpdateBody,
    TagView,
)
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.routes.api.dataset import get_and_validate_dataset_id
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CONFLICT,
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)
from lightly_purple.server.routes.api.validators import Paginated

tag_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_tag_resolver(session: SessionDep) -> TagResolver:
    """Create an instance of the TagResolver."""
    return TagResolver(session)


@tag_router.post(
    "/datasets/{dataset_id}/tags",
    response_model=TagView,
    status_code=HTTP_STATUS_CREATED,
)
def create_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    body: TagInputBody,
) -> Tag:
    """Create a new tag in the database."""
    dataset_id = dataset.dataset_id
    try:
        return handler.create(
            TagInput(
                **body.model_dump(exclude_unset=True), dataset_id=dataset_id
            )
        )
    except IntegrityError as e:
        raise HTTPException(
            status_code=HTTP_STATUS_CONFLICT,
            detail=f"""
                Tag with name {body.name} already exists
                in the dataset {dataset_id}.
            """,
        ) from e


@tag_router.get("/datasets/{dataset_id}/tags")
def read_tags(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    paginated: Annotated[Paginated, Query()],
) -> list[TagView]:
    """Retrieve a list of tags from the database."""
    return handler.get_all_by_dataset_id(
        **paginated.model_dump(),
        dataset_id=dataset.dataset_id,
    )


@tag_router.get("/datasets/{dataset_id}/tags/{tag_id}")
def read_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    tag_id: Annotated[UUID, Path(title="Tag Id")],
) -> Tag:
    """Retrieve a single tag from the database."""
    tag = handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"""
            Tag with id {tag_id} for dataset {dataset.dataset_id} not found.
            """,
        )
    return tag


@tag_router.put("/datasets/{dataset_id}/tags/{tag_id}")
def update_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    tag_id: Annotated[UUID, Path(title="Tag Id")],
    body: TagUpdateBody,
) -> Tag:
    """Update an existing tag in the database."""
    try:
        tag = handler.update(
            tag_id,
            TagUpdate(
                **body.model_dump(exclude_unset=True),
            ),
        )
        if not tag:
            raise HTTPException(
                status_code=HTTP_STATUS_NOT_FOUND,
                detail=f"Tag with id {tag_id} not found.",
            )
    except IntegrityError as e:
        raise HTTPException(
            status_code=HTTP_STATUS_CONFLICT,
            detail=f"""
                Cannot update tag. Tag with name {body.name}
                already exists in the dataset {dataset.dataset_id}.
            """,
        ) from e
    return tag


@tag_router.delete("/datasets/{dataset_id}/tags/{tag_id}")
def delete_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: Annotated[UUID, Path(title="Tag Id")],
):
    """Delete a tag from the database."""
    if not handler.delete(tag_id):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="tag not found"
        )
    return {"status": "deleted"}


class SampleIdsBody(BaseModel):
    """body parameters for adding or removing thing_ids."""

    sample_ids: list[UUID] | None = Field(
        None, description="sample ids to add/remove"
    )


@tag_router.post(
    "/datasets/{dataset_id}/tags/{tag_id}/add/samples",
    status_code=HTTP_STATUS_CREATED,
)
def add_sample_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: SampleIdsBody,
) -> bool:
    """Add sample_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't add sample_ids.",
        )

    tag_handler.add_sample_ids_to_tag_id(tag_id, **body.model_dump())
    return True


@tag_router.delete(
    "/datasets/{dataset_id}/tags/{tag_id}/remove/samples",
)
def remove_thing_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: SampleIdsBody,
) -> bool:
    """Add thing_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't remove samples.",
        )

    tag_handler.remove_sample_ids_from_tag_id(tag_id, **body.model_dump())
    return True


class AnnotationIdsBody(BaseModel):
    """body parameters for adding or removing annotation_ids."""

    annotation_ids: list[UUID] | None = Field(
        None, description="annotation ids to add/remove"
    )


@tag_router.post(
    "/datasets/{dataset_id}/tags/{tag_id}/add/annotations",
    status_code=HTTP_STATUS_CREATED,
)
def add_annotation_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: AnnotationIdsBody,
) -> bool:
    """Add thing_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't add annotations.",
        )

    tag_handler.add_annotation_ids_to_tag_id(tag_id, **body.model_dump())
    return True


@tag_router.delete(
    "/datasets/{dataset_id}/tags/{tag_id}/remove/annotations",
)
def remove_annotation_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: AnnotationIdsBody,
) -> bool:
    """Add thing_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't remove annotations.",
        )

    tag_handler.remove_annotation_ids_from_tag_id(tag_id, **body.model_dump())
    return True



"""API endpoints for annotation tasks."""

from typing import List
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, status
from sqlmodel import Session

from lightly_purple.server.db import get_session
from lightly_purple.server.models.annotation_task import AnnotationTask
from lightly_purple.server.resolvers.annotation_task import (
    AnnotationTaskResolver,
)

router = APIRouter(prefix="/annotationtasks", tags=["annotationtasks"])


@router.get("/", response_model=List[AnnotationTask])
def get_annotation_tasks(
    session: Session = Depends(get_session),  # noqa: B008
) -> List[AnnotationTask]:
    """Get all annotation tasks."""
    resolver = AnnotationTaskResolver(session)
    return resolver.get_all()


@router.get("/{annotation_task_id}", response_model=AnnotationTask)
def get_annotation_task(
    annotation_task_id: UUID,
    session: Session = Depends(get_session),  # noqa: B008
) -> AnnotationTask:
    """Get an annotation task by ID."""
    resolver = AnnotationTaskResolver(session)
    task = resolver.get_by_id(annotation_task_id)
    if task is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Annotation task with ID {annotation_task_id} not found",
        )
    return task



"""Handler for database operations related to datasets."""

from __future__ import annotations

from datetime import datetime, timezone
from uuid import UUID

from pydantic import BaseModel, Field, model_validator
from sqlmodel import Session, and_, func, or_, select

from lightly_purple.server.models import (
    BoundingBoxAnnotation,
    Dataset,
    Sample,
    Tag,
)
from lightly_purple.server.models.dataset import DatasetInput


class ExportFilter(BaseModel):
    """Export Filter to be used for including or excluding."""

    tag_ids: list[UUID] | None = Field(
        default=None, min_length=1, description="List of tag UUIDs"
    )
    sample_ids: list[UUID] | None = Field(
        default=None, min_length=1, description="List of sample UUIDs"
    )
    annotation_ids: list[UUID] | None = Field(
        default=None, min_length=1, description="List of annotation UUIDs"
    )

    @model_validator(mode="after")
    def check_exactly_one(cls, model: ExportFilter) -> ExportFilter:
        """Ensure that exactly one of the fields is set."""
        count = (
            (model.tag_ids is not None)
            + (model.sample_ids is not None)
            + (model.annotation_ids is not None)
        )
        if count != 1:
            raise ValueError(
                "Either tag_ids, sample_ids, or annotation_ids must be set."
            )
        return model


class DatasetResolver:
    """Resolver for the Dataset model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(self, dataset: DatasetInput) -> Dataset:
        """Create a new dataset in the database."""
        db_dataset = Dataset.model_validate(dataset)
        self.session.add(db_dataset)
        self.session.commit()
        self.session.refresh(db_dataset)
        return db_dataset

    def get_all(self, offset: int = 0, limit: int = 100) -> list[Dataset]:
        """Retrieve all datasets with pagination."""
        datasets = self.session.exec(
            select(Dataset).offset(offset).limit(limit)
        ).all()
        return list(datasets) if datasets else []

    def get_by_id(self, dataset_id: UUID) -> Dataset | None:
        """Retrieve a single dataset by ID."""
        return self.session.exec(
            select(Dataset).where(Dataset.dataset_id == dataset_id)
        ).one_or_none()

    def update(
        self, dataset_id: UUID, dataset_data: DatasetInput
    ) -> Dataset | None:
        """Update an existing dataset."""
        dataset = self.get_by_id(dataset_id)
        if not dataset:
            return None

        dataset.name = dataset_data.name
        dataset.directory = dataset_data.directory
        dataset.updated_at = datetime.now(timezone.utc)

        self.session.commit()
        self.session.refresh(dataset)
        return dataset

    def delete(self, dataset_id: UUID) -> bool:
        """Delete a dataset."""
        dataset = self.get_by_id(dataset_id)
        if not dataset:
            return False

        self.session.delete(dataset)
        self.session.commit()
        return True

    def _build_export_query(  # noqa: C901
        self,
        dataset_id: UUID,
        include: ExportFilter | None = None,
        exclude: ExportFilter | None = None,
    ) -> select:
        """Build the export query based on filters.

        Args:
            dataset_id: UUID of the dataset.
            include: Filter to include samples.
            exclude: Filter to exclude samples.

        Returns:
            SQLModel select query
        """
        if not include and not exclude:
            raise ValueError("Include or exclude filter is required.")
        if include and exclude:
            raise ValueError("Cannot include and exclude at the same time.")

        # include tags or sample_ids or annotation_ids from result
        if include:
            if include.tag_ids:
                return (
                    select(Sample)
                    .where(Sample.dataset_id == dataset_id)
                    .where(
                        or_(
                            # Samples with matching sample tags
                            Sample.tags.any(
                                and_(
                                    Tag.kind == "sample",
                                    Tag.tag_id.in_(include.tag_ids),
                                )
                            ),
                            # Samples with matching annotation tags
                            Sample.annotations.any(
                                BoundingBoxAnnotation.tags.any(
                                    and_(
                                        Tag.kind == "annotation",
                                        Tag.tag_id.in_(include.tag_ids),
                                    )
                                )
                            ),
                        )
                    )
                    .order_by(Sample.sample_id)
                    .distinct()
                )

            # get samples by specific sample_ids
            if include.sample_ids:
                return (
                    select(Sample)
                    .where(Sample.dataset_id == dataset_id)
                    .where(Sample.sample_id.in_(include.sample_ids))
                    .order_by(Sample.sample_id)
                    .distinct()
                )

            # get samples by specific annotation_ids
            if include.annotation_ids:
                return (
                    select(Sample)
                    .join(Sample.annotations)
                    .where(BoundingBoxAnnotation.dataset_id == dataset_id)
                    .where(
                        BoundingBoxAnnotation.annotation_id.in_(
                            include.annotation_ids
                        )
                    )
                    .order_by(Sample.sample_id)
                    .distinct()
                )

        # exclude tags or sample_ids or annotation_ids from result
        elif exclude:
            if exclude.tag_ids:
                return (
                    select(Sample)
                    .where(Sample.dataset_id == dataset_id)
                    .where(
                        and_(
                            ~Sample.tags.any(
                                and_(
                                    Tag.kind == "sample",
                                    Tag.tag_id.in_(exclude.tag_ids),
                                )
                            ),
                            or_(
                                ~Sample.annotations.any(),
                                ~Sample.annotations.any(
                                    BoundingBoxAnnotation.tags.any(
                                        and_(
                                            Tag.kind == "annotation",
                                            Tag.tag_id.in_(exclude.tag_ids),
                                        )
                                    )
                                ),
                            ),
                        )
                    )
                    .order_by(Sample.sample_id)
                    .distinct()
                )
            if exclude.sample_ids:
                return (
                    select(Sample)
                    .where(Sample.dataset_id == dataset_id)
                    .where(Sample.sample_id.notin_(exclude.sample_ids))
                    .order_by(Sample.sample_id)
                    .distinct()
                )
            if exclude.annotation_ids:
                return (
                    select(Sample)
                    .where(Sample.dataset_id == dataset_id)
                    .where(
                        or_(
                            ~Sample.annotations.any(),
                            ~Sample.annotations.any(
                                BoundingBoxAnnotation.annotation_id.in_(
                                    exclude.annotation_ids
                                )
                            ),
                        )
                    )
                    .order_by(Sample.sample_id)
                    .distinct()
                )

        raise ValueError("Invalid include or export filter combination.")

    # TODO: this fn should be moved to a "business logic" layer outside of the
    # resolvers and abstracted in to reusable components.
    # https://linear.app/lightly/issue/LIG-6196/figure-out-architecture-follow-up-changes-in-python-package
    # TODO: this should be abstracted to allow different export formats.
    def export(
        self,
        dataset_id: UUID,
        include: ExportFilter | None = None,
        exclude: ExportFilter | None = None,
    ) -> list[str]:
        """Retrieve samples for exporting from a dataset.

        Only one of include or exclude should be set and not both.
        Furthermore, the include and exclude filter can only have
        one type (tag_ids, sample_ids or annotations_ids) set.

        Args:
            dataset_id: UUID of the dataset.
            include: Filter to include samples.
            exclude: Filter to exclude samples.

        Returns:
            List of file paths
        """
        query = self._build_export_query(dataset_id, include, exclude)
        result = self.session.exec(query).all()
        return [sample.file_path_abs for sample in result]

    def get_filtered_samples_count(
        self,
        dataset_id: UUID,
        include: ExportFilter | None = None,
        exclude: ExportFilter | None = None,
    ) -> int:
        """Get statistics about the export query.

        Only one of include or exclude should be set and not both.
        Furthermore, the include and exclude filter can only have
        one type (tag_ids, sample_ids or annotations_ids) set.

        Args:
            dataset_id: UUID of the dataset.
            include: Filter to include samples.
            exclude: Filter to exclude samples.

        Returns:
            Count of files to be exported
        """
        query = self._build_export_query(dataset_id, include, exclude)
        count_query = select(func.count()).select_from(query.subquery())
        return self.session.exec(count_query).one() or 0



"""This module contains the API routes for managing datasets."""

from __future__ import annotations

from datetime import datetime, timezone
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Path, Query
from fastapi.responses import PlainTextResponse
from pydantic import BaseModel
from sqlmodel import Field, Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import Dataset
from lightly_purple.server.models.dataset import DatasetInput, DatasetView
from lightly_purple.server.resolvers.dataset import (
    DatasetResolver,
    ExportFilter,
)
from lightly_purple.server.routes.api.status import HTTP_STATUS_NOT_FOUND
from lightly_purple.server.routes.api.validators import Paginated

dataset_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_dataset_resolver(session: SessionDep) -> DatasetResolver:
    """Create an instance of the DatasetResolver."""
    return DatasetResolver(session)


def get_and_validate_dataset_id(
    dataset_handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    dataset_id: UUID,
) -> Dataset:
    """Get and validate the existence of a dataset on a route."""
    dataset = dataset_handler.get_by_id(dataset_id)
    if not dataset:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f""" Dataset with {dataset_id} not found.""",
        )
    return dataset


@dataset_router.post(
    "/datasets",
    response_model=DatasetView,
    status_code=201,
)
def create_dataset(
    dataset_input: DatasetInput,
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
) -> Dataset:
    """Create a new dataset in the database."""
    return handler.create(dataset_input)


@dataset_router.get("/datasets")
def read_datasets(
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    paginated: Annotated[Paginated, Query()],
) -> list[DatasetView]:
    """Retrieve a list of datasets from the database."""
    return handler.get_all(**paginated.model_dump())


@dataset_router.get("/datasets/{dataset_id}")
def read_dataset(
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
) -> Dataset:
    """Retrieve a single dataset from the database."""
    return dataset


@dataset_router.put("/datasets/{dataset_id}")
def update_dataset(
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    dataset_input: DatasetInput,
) -> Dataset:
    """Update an existing dataset in the database."""
    return handler.update(dataset.dataset_id, dataset_input)


@dataset_router.delete("/datasets/{dataset_id}")
def delete_dataset(
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
):
    """Delete a dataset from the database."""
    handler.delete(dataset.dataset_id)
    return {"status": "deleted"}


class ExportBody(BaseModel):
    """body parameters for including or excluding tag_ids or sample_ids."""

    include: ExportFilter | None = Field(
        None, description="include filter for sample_ids or tag_ids"
    )
    exclude: ExportFilter | None = Field(
        None, description="exclude filter for sample_ids or tag_ids"
    )


# This endpoint should be a GET, however due to the potential huge size
# of sample_ids, it is a POST request to avoid URL length limitations.
# A body with a GET request is supported by fastAPI however it has undefined
# behavior: https://fastapi.tiangolo.com/tutorial/body/
@dataset_router.post(
    "/datasets/{dataset_id}/export",
)
def export_dataset_to_absolute_paths(
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    body: ExportBody,
):
    """Export dataset from the database."""
    # export dataset to absolute paths
    exported = handler.export(
        dataset_id=dataset.dataset_id,
        include=body.include,
        exclude=body.exclude,
    )

    # Create a response with the exported data
    response = PlainTextResponse("\n".join(exported))

    # Add the Content-Disposition header to force download
    filename = f"{dataset.name}_exported_{datetime.now(timezone.utc)}.txt"
    response.headers["Access-Control-Expose-Headers"] = "Content-Disposition"
    response.headers["Content-Disposition"] = f"attachment; filename={filename}"

    return response


"""
Endpoint to export samples from a dataset.
"""


@dataset_router.post(
    "/datasets/{dataset_id}/export/stats",
)
def export_dataset_stats(
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    body: ExportBody,
) -> int:
    """Get statistics about the export query."""
    return handler.get_filtered_samples_count(
        dataset_id=dataset.dataset_id,
        include=body.include,
        exclude=body.exclude,
    )



"""This module contains the API routes for computing detection metrics."""

from __future__ import annotations

from uuid import UUID

from fastapi import APIRouter, Depends
from pydantic import BaseModel
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.metrics.detection.map import (
    DetectionMetricsMAP,
    calculate_map_metric,
)
from lightly_purple.server.db import get_session
from lightly_purple.server.resolvers.annotation import (
    AnnotationsFilterParams,
    BoundingBoxAnnotationResolver,
)
from lightly_purple.server.resolvers.annotation_label import (
    AnnotationLabelResolver,
)
from lightly_purple.server.routes.api.annotation import get_annotation_resolver
from lightly_purple.server.routes.api.annotation_label import (
    get_annotation_label_resolver,
)

metrics_router = APIRouter()

SessionDep = Annotated[Session, Depends(get_session)]


class DetectionMetricsMAPRequest(BaseModel):
    """Request for computing the MAP detection metric."""

    dataset_id: UUID
    ground_truth_task_id: UUID
    prediction_task_id: UUID
    tag_id: UUID | None = None


@metrics_router.post(
    "/metrics/compute/detection/map", response_model=DetectionMetricsMAP
)
def compute_detection_map(
    request_body: DetectionMetricsMAPRequest,
    resolver: Annotated[
        BoundingBoxAnnotationResolver, Depends(get_annotation_resolver)
    ],
    label_resolver: Annotated[
        AnnotationLabelResolver, Depends(get_annotation_label_resolver)
    ],
) -> DetectionMetricsMAP:
    """Compute the MAP detection metric."""
    ground_truth_annotations = resolver.get_all(
        filters=AnnotationsFilterParams(
            dataset_ids=[request_body.dataset_id],
            annotation_task_ids=[request_body.ground_truth_task_id],
            sample_tag_ids=[request_body.tag_id]
            if request_body.tag_id
            else None,
        ),
    )
    prediction_annotations = resolver.get_all(
        filters=AnnotationsFilterParams(
            dataset_ids=[request_body.dataset_id],
            annotation_task_ids=[request_body.prediction_task_id],
            sample_tag_ids=[request_body.tag_id]
            if request_body.tag_id
            else None,
        ),
    )

    metrics_result = calculate_map_metric(
        pred_annotations=prediction_annotations,
        gt_annotations=ground_truth_annotations,
    )

    # Rename per-class metrics to use label names
    raw_map_pc = metrics_result.map_per_class
    if raw_map_pc:
        id2name = label_resolver.names_by_ids([UUID(k) for k in raw_map_pc])
        metrics_result.map_per_class = {
            id2name.get(k, k): v for k, v in raw_map_pc.items()
        }
    raw_mar100_pc = metrics_result.mar_100_per_class
    if raw_mar100_pc:
        id2name = label_resolver.names_by_ids([UUID(k) for k in raw_mar100_pc])
        metrics_result.mar_100_per_class = {
            id2name.get(k, k): v for k, v in raw_mar100_pc.items()
        }
    return metrics_result



"""Dataset functionality module."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from uuid import UUID

from labelformat.formats import (
    COCOInstanceSegmentationInput,
    COCOObjectDetectionInput,
    YOLOv8ObjectDetectionInput,
)
from labelformat.model.binary_mask_segmentation import BinaryMaskSegmentation
from labelformat.model.bounding_box import BoundingBoxFormat
from labelformat.model.image import Image
from labelformat.model.instance_segmentation import (
    ImageInstanceSegmentation,
    InstanceSegmentationInput,
)
from labelformat.model.multipolygon import MultiPolygon
from labelformat.model.object_detection import (
    ImageObjectDetection,
    ObjectDetectionInput,
)
from tqdm import tqdm

from lightly_purple.dataset.embedding_generator import EmbeddingGenerator
from lightly_purple.dataset.embedding_manager import EmbeddingManagerProvider
from lightly_purple.dataset.env import APP_URL, PURPLE_HOST, PURPLE_PORT
from lightly_purple.server.db import db_manager
from lightly_purple.server.features import purple_active_features
from lightly_purple.server.models import Dataset
from lightly_purple.server.models.annotation_label import AnnotationLabelInput
from lightly_purple.server.models.annotation_task import (
    AnnotationTask,
    AnnotationType,
)
from lightly_purple.server.models.bounding_box_annotation import (
    BoundingBoxAnnotationInput,
)
from lightly_purple.server.models.dataset import DatasetInput
from lightly_purple.server.models.sample import SampleInput
from lightly_purple.server.models.tag import TagInput
from lightly_purple.server.resolvers.annotation import (
    BoundingBoxAnnotationResolver,
)
from lightly_purple.server.resolvers.annotation_label import (
    AnnotationLabelResolver,
)
from lightly_purple.server.resolvers.annotation_task import (
    AnnotationTaskResolver,
)
from lightly_purple.server.resolvers.dataset import DatasetResolver
from lightly_purple.server.resolvers.sample import SampleResolver
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.server import Server

# Constants
ANNOTATION_BATCH_SIZE = 64  # Number of annotations to process in a single batch
EMBEDDING_BATCH_SIZE = 64  # Number of embeddings to process in a single batch


@dataclass
class AnnotationProcessingContext:
    """Context for processing annotations for a single sample."""

    dataset_id: UUID
    sample_id: UUID
    label_map: dict[int, UUID]
    annotation_task_id: UUID


class DatasetLoader:
    """Class responsible for loading datasets from various sources."""

    def __init__(self) -> None:
        """Initialize the dataset loader."""
        with db_manager.session() as session:
            self.dataset_resolver = DatasetResolver(session)
            self.tag_resolver = TagResolver(session)
            self.sample_resolver = SampleResolver(session)
            self.annotation_resolver = BoundingBoxAnnotationResolver(session)
            self.embedding_manager = (
                EmbeddingManagerProvider.get_embedding_manager(session=session)
            )
            self.annotation_label_resolver = AnnotationLabelResolver(session)
            self.annotation_task_resolver = AnnotationTaskResolver(session)

    def _create_dataset(self, name: str, directory: str) -> Dataset:
        """Creates a new dataset."""
        with db_manager.session() as session:  # noqa: F841
            # Create dataset record
            dataset = DatasetInput(
                name=name,
                directory=directory,
            )
            return self.dataset_resolver.create(dataset)

    def _create_example_tags(self, dataset: Dataset) -> None:
        """Create example tags for samples and annotations."""
        self.tag_resolver.create(
            TagInput(
                dataset_id=dataset.dataset_id,
                name="label_mistakes",
                kind="sample",
            )
        )
        self.tag_resolver.create(
            TagInput(
                dataset_id=dataset.dataset_id,
                name="label_mistakes",
                kind="annotation",
            )
        )

    def _create_label_map(
        self, input_labels: ObjectDetectionInput | InstanceSegmentationInput
    ) -> dict[int, UUID]:
        """Create a mapping of category IDs to annotation label IDs."""
        label_map = {}
        for category in tqdm(
            input_labels.get_categories(),
            desc="Processing categories",
            unit=" categories",
        ):
            label = AnnotationLabelInput(annotation_label_name=category.name)
            stored_label = self.annotation_label_resolver.create(label)
            label_map[category.id] = stored_label.annotation_label_id
        return label_map

    def _process_object_detection_annotations(
        self,
        context: AnnotationProcessingContext,
        image_data: ImageObjectDetection,
    ) -> list[BoundingBoxAnnotationInput]:
        """Process object detection annotations for a single image."""
        new_annotations = []
        for obj in image_data.objects:
            box = obj.box.to_format(BoundingBoxFormat.XYWH)
            x, y, width, height = box

            new_annotations.append(
                BoundingBoxAnnotationInput(
                    dataset_id=context.dataset_id,
                    sample_id=context.sample_id,
                    annotation_label_id=context.label_map[obj.category.id],
                    x=x,
                    y=y,
                    width=width,
                    height=height,
                    confidence=obj.confidence,
                    annotation_task_id=context.annotation_task_id,
                )
            )
        return new_annotations

    def _process_instance_segmentation_annotations(
        self,
        context: AnnotationProcessingContext,
        image_data: ImageInstanceSegmentation,
    ) -> list[BoundingBoxAnnotationInput]:
        """Process instance segmentation annotations for a single image."""
        new_annotations = []
        for obj in image_data.objects:
            segmentation_rle: None | list[int] = None
            if isinstance(obj.segmentation, MultiPolygon):
                box = obj.segmentation.bounding_box().to_format(
                    BoundingBoxFormat.XYWH
                )
            elif isinstance(obj.segmentation, BinaryMaskSegmentation):
                box = obj.segmentation.bounding_box.to_format(
                    BoundingBoxFormat.XYWH
                )
                segmentation_rle = obj.segmentation._rle_row_wise  # noqa: SLF001
            else:
                raise ValueError(
                    f"Unsupported segmentation type: {type(obj.segmentation)}"
                )

            x, y, width, height = box

            new_annotations.append(
                BoundingBoxAnnotationInput(
                    dataset_id=context.dataset_id,
                    sample_id=context.sample_id,
                    annotation_label_id=context.label_map[obj.category.id],
                    x=x,
                    y=y,
                    width=width,
                    height=height,
                    segmentation__binary_mask__rle_row_wise=segmentation_rle,
                    annotation_task_id=context.annotation_task_id,
                )
            )
        return new_annotations

    def _load_into_dataset(
        self,
        dataset: Dataset,
        input_labels: ObjectDetectionInput | InstanceSegmentationInput,
        img_dir: Path,
        annotation_task_id: UUID,
    ) -> None:
        """Store a loaded dataset in database."""
        # Create label mapping
        label_map = self._create_label_map(input_labels)

        # Create example tags.
        self._create_example_tags(dataset=dataset)

        annotations_to_create: list[BoundingBoxAnnotationInput] = []
        sample_ids_for_embeddings = []

        # Load an embedding generator and register the model.
        embedding_generator = _load_embedding_generator()

        # Set feature flag if embedding_generator is available.
        if embedding_generator:
            purple_active_features.append("embeddingSearchEnabled")
            embedding_model = self.embedding_manager.register_embedding_model(
                dataset_id=dataset.dataset_id,
                embedding_generator=embedding_generator,
            )

        # Process images and annotations
        for image_data in tqdm(
            input_labels.get_labels(),
            desc="Processing images",
            unit=" images",
        ):
            # Mypy does not get that .image exists in both cases.
            image: Image = image_data.image  # type: ignore[attr-defined]
            # Create sample record
            sample = SampleInput(
                file_name=str(image.filename),
                file_path_abs=str(img_dir / image.filename),
                width=image.width,
                height=image.height,
                dataset_id=dataset.dataset_id,
            )
            stored_sample = self.sample_resolver.create(sample)
            sample_ids_for_embeddings.append(stored_sample.sample_id)
            # Process embedding batch if needed
            if (
                embedding_generator
                and len(sample_ids_for_embeddings) >= EMBEDDING_BATCH_SIZE
            ):
                self.embedding_manager.embed_images(
                    sample_ids=sample_ids_for_embeddings,
                    embedding_model_id=embedding_model.embedding_model_id,
                )
                sample_ids_for_embeddings = []

            # Create context for processing annotations for this sample
            context = AnnotationProcessingContext(
                dataset_id=dataset.dataset_id,
                sample_id=stored_sample.sample_id,
                label_map=label_map,
                annotation_task_id=annotation_task_id,
            )

            # Process annotations.
            new_annotations: list[BoundingBoxAnnotationInput] = []
            if isinstance(image_data, ImageInstanceSegmentation):
                new_annotations = (
                    self._process_instance_segmentation_annotations(
                        context=context,
                        image_data=image_data,
                    )
                )
            elif isinstance(image_data, ImageObjectDetection):
                new_annotations = self._process_object_detection_annotations(
                    context=context,
                    image_data=image_data,
                )
            else:
                raise ValueError(
                    f"Unsupported annotation type: {type(image_data)}"
                )

            annotations_to_create.extend(new_annotations)

            if len(annotations_to_create) >= ANNOTATION_BATCH_SIZE:
                self.annotation_resolver.create_many(annotations_to_create)
                annotations_to_create = []

        # Insert any remaining embeddings
        if embedding_generator and sample_ids_for_embeddings:
            self.embedding_manager.embed_images(
                sample_ids=sample_ids_for_embeddings,
                embedding_model_id=embedding_model.embedding_model_id,
            )
            sample_ids_for_embeddings = []

        # Insert any remaining annotations
        if annotations_to_create:
            self.annotation_resolver.create_many(annotations_to_create)

    def from_yolo(
        self,
        data_yaml_path: str,
        input_split: str = "train",
        task_name: str | None = None,
    ) -> None:
        """Load a dataset in YOLO format and store in DB.

        Args:
            data_yaml_path: Path to the YOLO data.yaml file.
            input_split: The split to load (e.g., 'train', 'val').
            task_name: Optional name for the annotation task. If None, a
                default name is generated.
        """
        data_yaml = Path(data_yaml_path).absolute()
        dataset_name = data_yaml.parent.name

        if task_name is None:
            task_name = (
                f"Loaded from YOLO: {data_yaml.name} ({input_split} split)"
            )

        # Load the dataset using labelformat.
        label_input = YOLOv8ObjectDetectionInput(
            input_file=data_yaml,
            input_split=input_split,
        )
        img_dir = label_input._images_dir()  # noqa: SLF001

        self.from_labelformat(
            input_labels=label_input,
            dataset_name=dataset_name,
            input_images_folder=str(img_dir),
            is_prediction=False,
            task_name=task_name,
        )

    def from_coco_object_detections(
        self,
        annotations_json_path: str,
        input_images_folder: str,
        task_name: str | None = None,
    ) -> None:
        """Load a dataset in COCO Object Detection format and store in DB.

        Args:
            annotations_json_path: Path to the COCO annotations JSON file.
            input_images_folder: Path to the folder containing the images.
            task_name: Optional name for the annotation task. If None, a
                default name is generated.
        """
        annotations_json = Path(annotations_json_path)
        dataset_name = annotations_json.parent.name

        if task_name is None:
            task_name = (
                f"Loaded from COCO Object Detection: {annotations_json.name}"
            )

        label_input = COCOObjectDetectionInput(
            input_file=annotations_json,
        )
        img_dir = Path(input_images_folder).absolute()

        self.from_labelformat(
            input_labels=label_input,
            dataset_name=dataset_name,
            input_images_folder=str(img_dir),
            is_prediction=False,
            task_name=task_name,
        )

    def from_coco_instance_segmentations(
        self,
        annotations_json_path: str,
        input_images_folder: str,
        task_name: str | None = None,
    ) -> None:
        """Load a dataset in COCO Instance Segmentation format and store in DB.

        Args:
            annotations_json_path: Path to the COCO annotations JSON file.
            input_images_folder: Path to the folder containing the images.
            task_name: Optional name for the annotation task. If None, a
                default name is generated.
        """
        annotations_json = Path(annotations_json_path)
        dataset_name = annotations_json.parent.name

        if task_name is None:
            task_name = f"Loaded from COCO Instance Segmentation: {annotations_json.name}"  # noqa: E501

        label_input = COCOInstanceSegmentationInput(
            input_file=annotations_json,
        )
        img_dir = Path(input_images_folder).absolute()

        self.from_labelformat(
            input_labels=label_input,
            dataset_name=dataset_name,
            input_images_folder=str(img_dir),
            is_prediction=False,
            task_name=task_name,
        )

    def from_labelformat(
        self,
        input_labels: ObjectDetectionInput | InstanceSegmentationInput,
        dataset_name: str,
        input_images_folder: str,
        is_prediction: bool = True,
        task_name: str | None = None,
    ) -> None:
        """Load a dataset from a labelformat object and store in database.

        Args:
            input_labels: The labelformat input object.
            dataset_name: The name for the new dataset.
            input_images_folder: Path to the folder containing the images.
            is_prediction: Whether the task is for prediction or labels.
            task_name: Optional name for the annotation task. If None, a
                default name is generated.
        """
        img_dir = Path(input_images_folder).absolute()

        # Determine annotation type based on input.
        # Currently, we always create BBOX tasks, even for segmentation,
        # as segmentation data is stored alongside bounding boxes.
        annotation_type = AnnotationType.BBOX

        # Generate a default task name if none is provided.
        if task_name is None:
            task_name = f"Loaded from labelformat: {dataset_name}"

        # Create dataset and annotation task.
        dataset = self._create_dataset(
            name=dataset_name,
            directory=str(img_dir),
        )
        new_annotation_task = self.annotation_task_resolver.create(
            AnnotationTask(
                name=task_name,
                annotation_type=annotation_type,
                is_prediction=is_prediction,
            )
        )

        img_dir = Path(input_images_folder).absolute()
        self._load_into_dataset(
            dataset=dataset,
            input_labels=input_labels,
            img_dir=img_dir,
            annotation_task_id=new_annotation_task.annotation_task_id,
        )

    def launch(self) -> None:
        """Launch the web interface for the loaded dataset."""
        server = Server(host=PURPLE_HOST, port=PURPLE_PORT)

        print(f"Open the Purple GUI under: {APP_URL}")

        server.start()


def _load_embedding_generator() -> EmbeddingGenerator | None:
    """Load the embedding generator.

    Use MobileCLIP if its dependencies have been installed,
    otherwise return None.
    """
    try:
        from lightly_purple.dataset.mobileclip_embedding_generator import (
            MobileCLIPEmbeddingGenerator,
        )

        print("Using MobileCLIP embedding generator.")
        return MobileCLIPEmbeddingGenerator()
    except ImportError:
        print("Embedding functionality is disabled.")
        return None



"""This module defines the BoundingBoxAnnotation model."""

from typing import List, Optional
from uuid import UUID

from sqlalchemy import ARRAY, Column, Integer
from sqlmodel import Field, SQLModel


class BoundingBoxAnnotationBase(SQLModel):
    """Base class for the BoundingBoxAnnotation model."""

    x: float
    y: float
    width: float
    height: float
    annotation_label_id: UUID = Field(
        foreign_key="annotation_labels.annotation_label_id"
    )
    dataset_id: UUID = Field(foreign_key="datasets.dataset_id")
    sample_id: UUID = Field(foreign_key="samples.sample_id")

    annotation_task_id: UUID = Field(
        foreign_key="annotation_tasks.annotation_task_id",
    )

    confidence: Optional[float] = None

    segmentation__binary_mask__rle_row_wise: Optional[List[int]] = Field(
        default=None,
        sa_column=Column(ARRAY(Integer), nullable=True),
    )


class BoundingBoxAnnotationInput(BoundingBoxAnnotationBase):
    """Bounding box annotation input model for creation."""



"""This module contains the API routes for managing annotations."""

from __future__ import annotations

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Path
from fastapi.params import Query
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import (
    AnnotationViewForAnnotationLabel,
    Dataset,
)
from lightly_purple.server.resolvers.annotation import (
    BoundingBoxAnnotationResolver,
)
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.routes.api.dataset import get_and_validate_dataset_id
from lightly_purple.server.routes.api.dataset_tag import get_tag_resolver
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)
from lightly_purple.server.routes.api.validators import Paginated

annotations_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_annotation_resolver(
    session: SessionDep,
) -> BoundingBoxAnnotationResolver:
    """Create an instance of the AnnotationResolver."""
    return BoundingBoxAnnotationResolver(session)


@annotations_router.get("/annotations/count")
def count_annotations_by_dataset(  # noqa: PLR0913 // FIXME: refactor to use proper pydantic
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    handler: Annotated[
        BoundingBoxAnnotationResolver, Depends(get_annotation_resolver)
    ],
    filtered_labels: Annotated[list[str] | None, Query()] = None,
    min_width: Annotated[int | None, Query(ge=0)] = None,
    max_width: Annotated[int | None, Query(ge=0)] = None,
    min_height: Annotated[int | None, Query(ge=0)] = None,
    max_height: Annotated[int | None, Query(ge=0)] = None,
    tag_ids: list[UUID] | None = None,
) -> list[dict[str, str | int]]:
    """Get annotation counts for a specific dataset.

    Returns a list of dictionaries with label name and count.
    """
    counts = handler.count_annotations_by_dataset(
        dataset_id=dataset.dataset_id,
        filtered_labels=filtered_labels,
        min_width=min_width,
        max_width=max_width,
        min_height=min_height,
        max_height=max_height,
        tag_ids=tag_ids,
    )
    return [
        {
            "label_name": label_name,
            "current_count": current_count,
            "total_count": total_count,
        }
        for label_name, current_count, total_count in counts
    ]


@annotations_router.get("/datasets/{dataset_id}/annotations")
def read_annotations(
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    handler: Annotated[
        BoundingBoxAnnotationResolver, Depends(get_annotation_resolver)
    ],
    pagination: Annotated[Paginated, Depends()],
    annotation_label_ids: Annotated[list[UUID] | None, Query()] = None,
    tag_ids: Annotated[list[UUID] | None, Query()] = None,
) -> list[AnnotationViewForAnnotationLabel]:
    """Retrieve a list of annotations from the database."""
    return handler.get_all(
        pagination={
            "offset": pagination.offset,
            "limit": pagination.limit,
        },
        filters={
            "dataset_ids": [dataset.dataset_id],
            "annotation_label_ids": annotation_label_ids or None,
            "tag_ids": tag_ids,
        },
    )


@annotations_router.post(
    "/annotations/{annotation_id}/tag/{tag_id}",
    status_code=HTTP_STATUS_CREATED,
)
def add_tag_to_annotation(
    annotation_handler: Annotated[
        BoundingBoxAnnotationResolver, Depends(get_annotation_resolver)
    ],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    annotation_id: UUID,
    tag_id: UUID,
) -> bool:
    """Add annotation to a tag."""
    annotation = annotation_handler.get_by_id(annotation_id)
    if not annotation:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Annotation {annotation_id} not found",
        )

    if not tag_handler.add_tag_to_annotation(tag_id, annotation):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True


@annotations_router.delete("/annotations/{annotation_id}/tag/{tag_id}")
def remove_tag_from_annotation(
    annotation_handler: Annotated[
        BoundingBoxAnnotationResolver, Depends(get_annotation_resolver)
    ],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    annotation_id: UUID,
) -> bool:
    """Remove annotation from a tag."""
    annotation = annotation_handler.get_by_id(annotation_id)
    if not annotation:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Annotation {annotation_id} not found",
        )

    if not tag_handler.remove_tag_from_annotation(tag_id, annotation):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True



"""This module contains the API routes for managing datasets."""

from pathlib import Path

from fastapi import APIRouter, HTTPException
from fastapi.responses import FileResponse

from .api.status import HTTP_STATUS_NOT_FOUND

app_router = APIRouter()

# Get the current project root directory
project_root = Path(__file__).resolve().parent.parent.parent

webapp_dir = project_root / "dist-purple-view-app"

# Check if the webapp directory exists and raise an error if it doesn't
if not webapp_dir.exists():
    raise RuntimeError(
        f"Directory '{webapp_dir}' does not exist in '{project_root}'"
    )

# Ensure the path is absolute
webapp_dir = webapp_dir.resolve()

# ensure the webapp index.html file exists
index_file = webapp_dir / "index.html"
if not index_file.exists():
    raise RuntimeError("No index file. Did you forget to build the webapp?")


@app_router.get("/{path:path}", include_in_schema=False)
async def serve_static_webapp_files_or_default_index_file(path: str):
    """Serve static files of webapp or serve the index.html file.

    Try to serve static files with file extensions or return 404.
    If no file extension, return the main webapp index.html.
    """
    file_path = webapp_dir / path

    # if file has an extension, try to return the file
    if file_path.suffix:
        if not file_path.exists() or not file_path.is_file():
            raise HTTPException(
                status_code=HTTP_STATUS_NOT_FOUND,
                detail=f"File '{path}' not found",
            )
        return FileResponse(file_path)

    # if file has no extension, return the index.html file regardless of path
    return FileResponse(index_file)



"""File manipulation utilities."""

from __future__ import annotations

import shutil
from pathlib import Path

import requests
import xxhash


def download_file_if_does_not_exist(url: str, local_filename: Path) -> None:
    """Download a file from a URL if it does not already exist locally."""
    if local_filename.exists():
        return
    with requests.get(url, stream=True) as r, open(local_filename, "wb") as f:
        shutil.copyfileobj(r.raw, f)


def get_file_xxhash(file_path: Path) -> str:
    """Calculate the xxhash of a file.

    XXHash is a fast non-cryptographic hash function.

    Args:
        file_path: Path to the file.

    Returns:
        The xxhash of the file as a string.
    """
    hasher = xxhash.xxh64()
    with file_path.open("rb") as f:
        while chunk := f.read(8192):
            hasher.update(chunk)
    return hasher.hexdigest()



"""This module defines the Embedding_Model model for the application."""

from __future__ import annotations

from uuid import UUID

from sqlmodel import CHAR, Column, Field, SQLModel


class EmbeddingModelBase(SQLModel):
    """Base class for the EmbeddingModel."""

    name: str
    parameter_count_in_mb: int | None = None
    embedding_model_hash: str | None = Field(
        default=None, sa_column=Column(CHAR(128))
    )
    embedding_dimension: int
    dataset_id: UUID = Field(default=None, foreign_key="datasets.dataset_id")


class EmbeddingModelInput(EmbeddingModelBase):
    """EmbeddingModel class when inserting."""



"""Setup logging for the project.

Assumed to be called before any other module is imported. Make sure no internal
modules are called from this file.

Note: In python, module content is loaded only once. Therefore we can safely
put the logic in the global scope.
"""

import logging
import warnings

# Set logging level to ERROR for labelformat.
logging.getLogger("labelformat").setLevel(logging.ERROR)

# Suppress warnings from mobileclip.
# TODO(Michal, 04/2025): Remove once we don't vendor mobileclip.
warnings.filterwarnings(
    "ignore", category=FutureWarning, module="timm.models.layers"
)
warnings.filterwarnings("ignore", category=FutureWarning, module="mobileclip")



"""EmbeddingGenerator implementations."""

from __future__ import annotations

import random
from pathlib import Path
from typing import Protocol, runtime_checkable
from uuid import UUID

from lightly_purple.server.models.embedding_model import EmbeddingModelInput


@runtime_checkable
class EmbeddingGenerator(Protocol):
    """Protocol defining the interface for embedding models.

    This protocol defines the interface that all embedding models must
    implement. Concrete implementations will use different techniques
    for creating embeddings.
    """

    def get_embedding_model_input(
        self, dataset_id: UUID
    ) -> EmbeddingModelInput:
        """Generate an EmbeddingModelInput instance.

        Args:
            dataset_id: The ID of the dataset.

        Returns:
            An EmbeddingModelInput instance with the model details.
        """

    def embed_text(self, text: str) -> list[float]:
        """Generate an embedding for a text sample.

        Args:
            text: The text to embed.

        Returns:
            A list of floats representing the generated embedding.
        """
        ...

    def embed_images(self, filepaths: list[Path]) -> list[list[float]]:
        """Generate embeddings for multiple image samples.

        TODO(Michal, 04/2025): Use DatasetLoader as input instead.

        Args:
            filepaths: A list of file paths to the images to embed.

        Returns:
            A list of lists of floats representing the generated embeddings
            in the same order as the input file paths.
        """
        ...


class RandomEmbeddingGenerator(EmbeddingGenerator):
    """Model that produces random embeddings with a fixed dimension."""

    def __init__(self, dimension: int = 3):
        """Initialize the random embedding model.

        Args:
            dimension: The dimension of the embedding vectors to generate.
        """
        self._dimension = dimension

    def get_embedding_model_input(
        self, dataset_id: UUID
    ) -> EmbeddingModelInput:
        """Generate an EmbeddingModelInput instance.

        Args:
            dataset_id: The ID of the dataset.

        Returns:
            An EmbeddingModelInput instance with the model details.
        """
        return EmbeddingModelInput(
            name="Random",
            embedding_model_hash="random_model",
            embedding_dimension=self._dimension,
            dataset_id=dataset_id,
        )

    def embed_text(self, _text: str) -> list[float]:
        """Generate a random embedding for a text sample."""
        return [random.random() for _ in range(self._dimension)]

    def embed_images(self, filepaths: list[Path]) -> list[list[float]]:
        """Generate random embeddings for multiple image samples."""
        return [
            [random.random() for _ in range(self._dimension)]
            for _ in range(len(filepaths))
        ]



"""Example: index a YOLO dataset **and** add model predictions.

Workflow
--------
1. Load the ground truth YOLO dataset (same loader call as `example.py`).
2. Create a second annotation task flagged `is_prediction=True`.
3. For every ground truth box we fabricate a “prediction” (copy the box,
   give it a random confidence); insert them in a single bulk call.
4. Compute mAP (overall + per class) in Python and print it.
5. Launch the Purple UI so you can inspect both tasks.

Environment
-----------
• DATASET_PATH: path to your YOLO `data.yaml`
• PURPLE_DATASET_SPLIT: dataset split to load (train / val / test)
"""

from __future__ import annotations

import os
import random
from pathlib import Path
from uuid import UUID

from lightly_purple.dataset.loader import DatasetLoader
from lightly_purple.metrics.detection.map import calculate_map_metric
from lightly_purple.server.db import db_manager
from lightly_purple.server.models.annotation_task import (
    AnnotationTask,
    AnnotationType,
)
from lightly_purple.server.models.bounding_box_annotation import (
    BoundingBoxAnnotationInput,
)
from lightly_purple.server.models.tag import TagInput
from lightly_purple.server.resolvers.annotation import (
    AnnotationsFilterParams,
    BoundingBoxAnnotationResolver,
)
from lightly_purple.server.resolvers.annotation_label import (
    AnnotationLabelResolver,
)
from lightly_purple.server.resolvers.annotation_task import (
    AnnotationTaskResolver,
)
from lightly_purple.server.resolvers.dataset import DatasetResolver
from lightly_purple.server.resolvers.sample import SampleResolver

# Tag imports for train/test split
from lightly_purple.server.resolvers.tag import TagResolver

# --------------------------------------------------------------------------- #
# Configuration
# --------------------------------------------------------------------------- #
DATASET_PATH = Path(
    os.getenv(
        "DATASET_PATH",
        "/path/to/your/yolo/dataset/data.yaml",
    )
).expanduser()
INPUT_SPLIT = os.getenv("PURPLE_DATASET_SPLIT", "train")  # train / val / test
CONF_RANGE = (0.5, 0.95)  # random confidence for mock predictions


# --------------------------------------------------------------------------- #
# Helpers
# --------------------------------------------------------------------------- #
def _random_confidence() -> float:
    return random.uniform(*CONF_RANGE)


# --------------------------------------------------------------------------- #
# Inference stub — replace with your own model call
# --------------------------------------------------------------------------- #
def _run_inference(
    image_files: list[Path],
) -> dict[str, list[tuple[float, float, float, float, float]]]:
    """Mock detector.

    Returns a mapping: image filename -> list of (x, y, w, h, confidence)

    Replace this with your actual batch inference implementation.
    The coordinates must be absolute pixel values (XYWH).
    """
    preds: dict[str, list[tuple[float, float, float, float, float]]] = {}
    for img in image_files:
        # fabricate 0 to 3 random boxes per image
        num_boxes = random.randint(0, 3)
        pred_boxes = []
        for _ in range(num_boxes):
            x = random.uniform(0, 400)
            y = random.uniform(0, 400)
            w = random.uniform(20, 150)
            h = random.uniform(20, 150)
            conf = _random_confidence()
            pred_boxes.append((x, y, w, h, conf))
        preds[img.name] = pred_boxes
    return preds


def _latest_dataset_id(resolver: DatasetResolver):
    """Return the newest dataset (assumes each run starts fresh)."""
    datasets = resolver.get_all()
    return datasets[0].dataset_id  # type: ignore[attr-defined]


def _latest_gt_task_id(resolver: AnnotationTaskResolver):
    """Return the newest *ground truth* task (is_prediction = False)."""
    for task in reversed(resolver.get_all()):
        if (
            not task.is_prediction
            and task.annotation_type == AnnotationType.BBOX
        ):
            return task.annotation_task_id
    raise RuntimeError("No ground truth task found.")


# --------------------------------------------------------------------------- #
# 1. Load ground truth YOLO labels
# --------------------------------------------------------------------------- #
print("▶ Loading YOLO dataset…")
loader = DatasetLoader()
loader.from_yolo(
    str(DATASET_PATH), input_split=INPUT_SPLIT, task_name="Labelbox Annotations"
)

# --------------------------------------------------------------------------- #
# 2. Open a DB session: create prediction task & insert fake boxes
# --------------------------------------------------------------------------- #
with db_manager.session() as session:
    dataset_resolver = DatasetResolver(session)
    task_resolver = AnnotationTaskResolver(session)
    ann_resolver = BoundingBoxAnnotationResolver(session)

    dataset_id = _latest_dataset_id(dataset_resolver)
    gt_task_id = _latest_gt_task_id(task_resolver)
    # --- fetch all GT boxes (needed for metrics & label pool) ------------- #
    gt_boxes = ann_resolver.get_all(
        filters=AnnotationsFilterParams(annotation_task_ids=[gt_task_id])
    )

    # Use all distinct label ids from the GT set
    label_id_pool = list({box.annotation_label_id for box in gt_boxes})

    # --- create prediction task ------------------------------------------- #
    pred_task = task_resolver.create(
        AnnotationTask(
            name="YOLOv8 predictions",
            annotation_type=AnnotationType.BBOX,
            is_prediction=True,
        )
    )
    pred_task_id = pred_task.annotation_task_id
    print(f"▶ Created prediction task {pred_task_id}")

    # --- gather sample records to map filename -> sample_id --------------- #
    sample_resolver = SampleResolver(session)
    # Fetch *all* samples for this dataset (single call with large limit)
    samples = sample_resolver.get_all_by_dataset_id(
        dataset_id=dataset_id,
        offset=0,
        limit=100_000,  # adjust if you expect >100k images
    )
    # Create train/test split tags (60/40) and assign samples
    tag_resolver = TagResolver(session)
    sample_ids = [s.sample_id for s in samples]
    random.shuffle(sample_ids)
    split_idx = int(0.6 * len(sample_ids))
    train_ids = sample_ids[:split_idx]
    test_ids = sample_ids[split_idx:]

    train_tag = tag_resolver.create(
        TagInput(dataset_id=dataset_id, name="train", kind="sample")
    )
    test_tag = tag_resolver.create(
        TagInput(dataset_id=dataset_id, name="test", kind="sample")
    )
    tag_resolver.add_sample_ids_to_tag_id(train_tag.tag_id, train_ids)
    tag_resolver.add_sample_ids_to_tag_id(test_tag.tag_id, test_ids)
    filename_to_sample: dict[str, UUID] = {}
    for s in samples:
        # Sample model stores the absolute path in `file_path_abs`
        fname = Path(getattr(s, "file_path_abs", s.file_name)).name
        filename_to_sample[fname] = s.sample_id

    # --- call user defined inference -------------------------------------- #
    image_paths = [
        Path(getattr(s, "file_path_abs", s.file_name)) for s in samples
    ]
    pred_dict = _run_inference(image_paths)

    # --- convert predictions to DB inputs --------------------------------- #
    pred_inputs: list[BoundingBoxAnnotationInput] = []
    for fname, boxes in pred_dict.items():
        sample_id = filename_to_sample.get(fname)
        if sample_id is None:
            continue  # skip files that are not in our DB
        for x, y, w, h, conf in boxes:
            pred_inputs.append(
                BoundingBoxAnnotationInput(
                    dataset_id=dataset_id,
                    sample_id=sample_id,
                    annotation_label_id=random.choice(label_id_pool),
                    annotation_task_id=pred_task_id,
                    x=x,
                    y=y,
                    width=w,
                    height=h,
                    confidence=conf,
                )
            )

    # Add jittered ground-truth boxes as pseudo-predictions for some positives
    for gt in gt_boxes:
        if random.random() < 0.8:  # noqa: PLR2004
            pred_inputs.append(
                BoundingBoxAnnotationInput(
                    dataset_id=gt.dataset_id,
                    sample_id=gt.sample_id,
                    annotation_label_id=gt.annotation_label_id,
                    annotation_task_id=pred_task_id,
                    x=gt.x + random.uniform(-5, 5),
                    y=gt.y + random.uniform(-5, 5),
                    width=gt.width * random.uniform(0.9, 1.1),
                    height=gt.height * random.uniform(0.9, 1.1),
                    confidence=random.uniform(0.8, 1.0),
                )
            )

    ann_resolver.create_many(pred_inputs)
    print(f"▶ Inserted {len(pred_inputs)} predictions")

    # --------------------------------------------------------------------- #
    # 3. Compute metrics locally
    # --------------------------------------------------------------------- #
    metrics = calculate_map_metric(pred_inputs, gt_boxes)
    print("\n=== mAP ===")
    print(f"Overall: {metrics.map:.3f}")
    if metrics.map_per_class:
        id2name = AnnotationLabelResolver(session).names_by_ids(
            [UUID(k) for k in metrics.map_per_class]
        )
        for label_id, val in metrics.map_per_class.items():
            print(f"  {id2name.get(label_id, label_id)}: {val:.3f}")
    print(f"mAP@0.50 (IoU=0.50): {metrics.map_50:.3f}")
    print(f"mAP@0.75 (IoU=0.75): {metrics.map_75:.3f}")
    print(f"mAP small objects (<32^2 px): {metrics.map_small:.3f}")
    print(f"mAP medium objects (32^2-96^2 px): {metrics.map_medium:.3f}")
    print(f"mAP large objects (>96^2 px): {metrics.map_large:.3f}")
    print(f"Recall@100 detections per image: {metrics.mar_100:.3f}")
    print("============\n")

# --------------------------------------------------------------------------- #
# 4. Launch UI
# --------------------------------------------------------------------------- #
loader.launch()



# Set up logging before importing any other modules.
# Add noqa to silence unused import and unsorted imports linter warnings.
from . import setup_logging  # noqa: F401 I001

from lightly_purple.dataset.loader import DatasetLoader


__all__ = ["DatasetLoader"]



"""Implementation for calculating Mean Average Precision (MAP)."""

from __future__ import annotations

import math
from collections import defaultdict
from collections.abc import Sequence
from uuid import UUID

import torch
from pydantic import BaseModel
from torch import Tensor
from torchmetrics.detection import MeanAveragePrecision

from lightly_purple.server.models.bounding_box_annotation import (
    BoundingBoxAnnotationBase,
)


class DetectionMetricsMAP(BaseModel):
    """Response for computing the MAP detection metric."""

    name: str
    map: float
    map_small: float
    map_medium: float
    map_large: float
    mar_1: float
    mar_10: float
    mar_100: float
    mar_small: float
    mar_medium: float
    mar_large: float
    map_50: float
    map_75: float
    map_per_class: dict[str, float] | None = None
    mar_100_per_class: dict[str, float] | None = None
    classes: list[int]

    @property
    def value(self) -> float:
        """Backwards compatibility alias for map."""
        return self.map

    @property
    def per_class(self) -> dict[str, float] | None:
        """Backwards compatibility alias for map_per_class."""
        return self.map_per_class


def calculate_map_metric(  # noqa: C901
    pred_annotations: Sequence[BoundingBoxAnnotationBase],
    gt_annotations: Sequence[BoundingBoxAnnotationBase],
) -> DetectionMetricsMAP:
    """Calculate the Mean Average Precision (MAP) metric."""
    if not gt_annotations or not pred_annotations:
        return DetectionMetricsMAP(
            name="mAP@.5:.95",
            map=0.0,
            map_small=0.0,
            map_medium=0.0,
            map_large=0.0,
            mar_1=0.0,
            mar_10=0.0,
            mar_100=0.0,
            mar_small=0.0,
            mar_medium=0.0,
            mar_large=0.0,
            map_50=-1.0,
            map_75=-1.0,
            map_per_class=None,
            mar_100_per_class=None,
            classes=[],
        )
    uuid_to_int_label: dict[UUID, int] = {}

    # Map sample_id to annotations.
    sample_id_to_annotations = _group_by_sample_id(
        pred_annotations=pred_annotations,
        gt_annotations=gt_annotations,
    )

    # Create the MeanAveragePrecision object.
    map_metric = MeanAveragePrecision(
        box_format="xywh",
        iou_thresholds=None,
        class_metrics=True,
        average="macro",
        backend="faster_coco_eval",
    )

    # Compute MAP updating the object sample by sample.
    for _sample_id, (
        sample_pred_annotations,
        sample_gt_annotations,
    ) in sample_id_to_annotations.items():
        # Convert to torchmetrics format.
        prediction_tm = _convert_to_torchmetrics(
            annotations=sample_pred_annotations,
            is_prediction=True,
            uuid_to_int_label_map=uuid_to_int_label,
        )
        ground_truth_tm = _convert_to_torchmetrics(
            annotations=sample_gt_annotations,
            is_prediction=False,
            uuid_to_int_label_map=uuid_to_int_label,
        )

        # Update the metric.
        map_metric.update(preds=[prediction_tm], target=[ground_truth_tm])

    # Compute the final results.
    results = map_metric.compute()
    # Invert label map to recover original UUIDs
    int_to_uuid = {v: k for k, v in uuid_to_int_label.items()}

    # Helper to convert Tensor to Python list
    def to_list(t: Tensor) -> list[float]:
        if t.dim() == 0:
            return [t.item()]
        return t.tolist()

    # Scalar metrics
    map_val = results["map"].item()
    map_small = results["map_small"].item()
    map_medium = results["map_medium"].item()
    map_large = results["map_large"].item()
    mar_1 = results["mar_1"].item()
    mar_10 = results["mar_10"].item()
    mar_100 = results["mar_100"].item()
    mar_small = results["mar_small"].item()
    mar_medium = results["mar_medium"].item()
    mar_large = results["mar_large"].item()

    # IoU-specific metrics
    map_50 = results.get("map_50")
    map_50 = (
        map_50.item()
        if isinstance(map_50, Tensor) and map_50.dim() == 0
        else -1.0
    )
    map_75 = results.get("map_75")
    map_75 = (
        map_75.item()
        if isinstance(map_75, Tensor) and map_75.dim() == 0
        else -1.0
    )

    # Per-class metrics
    raw_map_pc = results.get("map_per_class")
    per_class_dict: dict[str, float] | None = None
    if raw_map_pc is not None:
        pc_list = to_list(raw_map_pc)
        per_class_dict = {}
        for idx, val in enumerate(pc_list):
            if not math.isnan(val):
                per_class_dict[str(int_to_uuid[idx])] = val

    # Per-class recall at 100 detections
    raw_mar100_pc = results.get("mar_100_per_class")
    mar_100_pc_dict: dict[str, float] | None = None
    if raw_mar100_pc is not None:
        m100_list = to_list(raw_mar100_pc)
        mar_100_pc_dict = {}
        for idx, val in enumerate(m100_list):
            if not math.isnan(val):
                mar_100_pc_dict[str(int_to_uuid[idx])] = val

    # Observed classes
    raw_classes = results.get("classes")
    classes_list_output: list[int] = []
    if raw_classes is not None:
        classes_list_output = [int(x) for x in to_list(raw_classes)]

    return DetectionMetricsMAP(
        name="mAP@.5:.95",
        map=map_val,
        map_small=map_small,
        map_medium=map_medium,
        map_large=map_large,
        mar_1=mar_1,
        mar_10=mar_10,
        mar_100=mar_100,
        mar_small=mar_small,
        mar_medium=mar_medium,
        mar_large=mar_large,
        map_50=map_50,
        map_75=map_75,
        map_per_class=per_class_dict,
        mar_100_per_class=mar_100_pc_dict,
        classes=classes_list_output,
    )


def _group_by_sample_id(
    pred_annotations: Sequence[BoundingBoxAnnotationBase],
    gt_annotations: Sequence[BoundingBoxAnnotationBase],
) -> dict[
    UUID,
    tuple[list[BoundingBoxAnnotationBase], list[BoundingBoxAnnotationBase]],
]:
    """Group prediction and ground truth annotations by sample_id.

    Returns a dictionary with sample_id as key and a tuple of
    (list of prediction annotations, list of ground truth annotations) as value.
    """
    sample_id_to_annotations: defaultdict[
        UUID,
        tuple[list[BoundingBoxAnnotationBase], list[BoundingBoxAnnotationBase]],
    ] = defaultdict(lambda: ([], []))
    for ann in pred_annotations:
        sample_id_to_annotations[ann.sample_id][0].append(ann)
    for ann in gt_annotations:
        sample_id_to_annotations[ann.sample_id][1].append(ann)
    return sample_id_to_annotations


def _convert_to_torchmetrics(
    annotations: Sequence[BoundingBoxAnnotationBase],
    is_prediction: bool,
    uuid_to_int_label_map: dict[UUID, int],
) -> dict[str, Tensor]:
    """Convert annotations to torchmetrics format.

    Args:
        annotations: List of bounding box annotations.
        is_prediction: Whether the annotations are predictions.
        uuid_to_int_label_map: Map from UUID to integer label.

    Returns:
        Dictionary in torchmetrics format. For predictions the keys are
        `boxes`, `scores`, `labels`. For ground truth they are `boxes`,
        `labels`.

    """
    if not annotations:
        empty_boxes = torch.empty((0, 4), dtype=torch.float32)
        empty_labels = torch.empty((0,), dtype=torch.int64)
        if is_prediction:
            return {
                "boxes": empty_boxes,
                "scores": torch.empty((0,), dtype=torch.float32),
                "labels": empty_labels,
            }
        return {"boxes": empty_boxes, "labels": empty_labels}

    boxes = torch.tensor(
        [[a.x, a.y, a.width, a.height] for a in annotations],
        dtype=torch.float32,
    )

    mapped_labels = []
    # Use the passed-in map instead of a global one
    for uuid_label in (a.annotation_label_id for a in annotations):
        if uuid_label not in uuid_to_int_label_map:
            # Assign the next available integer ID based on the current map size
            uuid_to_int_label_map[uuid_label] = len(uuid_to_int_label_map)
        mapped_labels.append(uuid_to_int_label_map[uuid_label])

    labels = torch.tensor(mapped_labels, dtype=torch.int64)

    if is_prediction:
        scores = torch.tensor(
            [a.confidence for a in annotations], dtype=torch.float32
        )
        return {
            "boxes": boxes,
            "scores": scores,
            "labels": labels,
        }
    return {"boxes": boxes, "labels": labels}



"""This module defines the SampleEmbedding model for the application."""

from __future__ import annotations

from uuid import UUID

from sqlalchemy import ARRAY, Float
from sqlmodel import Column, Field, SQLModel


class EmbeddingBase(SQLModel):
    """Base class for the Embedding."""

    embedding_model_id: UUID = Field(
        foreign_key="embedding_models.embedding_model_id", primary_key=True
    )
    embedding: list[float] = Field(sa_column=Column(ARRAY(Float)))


class SampleEmbeddingBase(EmbeddingBase):
    """Base class for the Embeddings used for Samples."""

    sample_id: UUID = Field(foreign_key="samples.sample_id", primary_key=True)


class SampleEmbeddingInput(SampleEmbeddingBase):
    """Sample class when inserting."""



"""This module contains the API routes for managing samples."""

from __future__ import annotations

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Path, Query
from pydantic import BaseModel, Field
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import (
    BoundingBoxAnnotation,
    Dataset,
    Sample,
    SampleView,
)
from lightly_purple.server.models.sample import SampleInput
from lightly_purple.server.resolvers.annotation import (
    BoundingBoxAnnotationResolver,
)
from lightly_purple.server.resolvers.sample import SampleResolver
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.routes.api.annotation import get_annotation_resolver
from lightly_purple.server.routes.api.dataset import get_and_validate_dataset_id
from lightly_purple.server.routes.api.dataset_tag import get_tag_resolver
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)
from lightly_purple.server.routes.api.validators import Paginated

samples_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_sample_resolver(session: SessionDep) -> SampleResolver:
    """Create an instance of the SampleResolver."""
    return SampleResolver(session)


@samples_router.post("/samples", response_model=SampleView)
def create_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    input_sample: SampleInput,
) -> Sample:
    """Create a new sample in the database."""
    return handler.create(input_sample)


# Define additional query parameters model
class SubqueryFilterParams(BaseModel):
    """Query parameters for filtering and doing slice and dice with samples."""

    # metadata
    min_width: int | None = Field(None, ge=0, description="Minimum width")
    max_width: int | None = Field(None, ge=0, description="Maximum width")
    min_height: int | None = Field(None, ge=0, description="Minimum height")
    max_height: int | None = Field(None, ge=0, description="Maximum height")

    # labels
    annotation_labels: list[str] | None = Field(
        None, description="Annotation labels"
    )

    # tags
    tag_ids: list[UUID] | None = Field(None, description="Tag IDs to filter by")

    # text embeddings search
    text_embedding: list[float] | None = Field(
        None, description="Text embedding to search for"
    )


# TODO: this should be prefixed with /datasets/{dataset_id}/samples
class ReadSamplesQuery(SubqueryFilterParams, Paginated):
    """Query parameters for reading samples."""

    dataset_id: UUID = Field(description="Dataset ID")


@samples_router.get("/samples")
def read_samples(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    query: Annotated[ReadSamplesQuery, Query()],
    _dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
) -> list[SampleView]:
    """Retrieve a list of samples from the database with optional filtering."""
    return handler.get_all_by_dataset_id(
        **query.model_dump(),
    )


@samples_router.get("/samples/dimensions")
def get_sample_dimensions(
    sample_handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    dataset: Annotated[
        Dataset, Path(title="Dataset Id"), Depends(get_and_validate_dataset_id)
    ],
    annotation_labels: Annotated[list[str] | None, Query()] = None,
) -> dict[str, int]:
    """Get min and max dimensions of samples in a dataset."""
    return sample_handler.get_dimension_bounds(
        dataset_id=dataset.dataset_id, annotation_labels=annotation_labels
    )


@samples_router.get("/samples/{sample_id}")
def read_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
) -> SampleView:
    """Retrieve a single sample from the database."""
    sample = handler.get_by_id(sample_id)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Sample not found"
        )

    return sample


@samples_router.get("/samples/{sample_id}/annotations")
def read_sample_annotations(
    handler: Annotated[
        BoundingBoxAnnotationResolver, Depends(get_annotation_resolver)
    ],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
) -> list[BoundingBoxAnnotation]:
    """Retrieve annotations for sample."""
    return handler.get_by_sample_id(sample_id)


@samples_router.put("/samples/{sample_id}")
def update_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
    sample_input: Sample,
) -> Sample:
    """Update an existing sample in the database."""
    sample = handler.update(sample_id, sample_input)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Sample not found"
        )
    return sample


@samples_router.delete("/samples/{sample_id}")
def delete_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
):
    """Delete a sample from the database."""
    if not handler.delete(sample_id):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Sample not found"
        )
    return {"status": "deleted"}


@samples_router.post(
    "/samples/{sample_id}/tag/{tag_id}",
    status_code=HTTP_STATUS_CREATED,
)
def add_tag_to_sample(
    sample_handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    sample_id: UUID,
    tag_id: UUID,
) -> bool:
    """Add sample to a tag."""
    sample = sample_handler.get_by_id(sample_id)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Sample {sample_id} not found",
        )

    if not tag_handler.add_tag_to_sample(tag_id, sample):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True


@samples_router.delete("/samples/{sample_id}/tag/{tag_id}")
def remove_tag_from_sample(
    sample_handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    sample_id: UUID,
) -> bool:
    """Remove sample from a tag."""
    sample = sample_handler.get_by_id(sample_id)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Sample {sample_id} not found",
        )

    if not tag_handler.remove_tag_from_sample(tag_id, sample):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True



"""This module contains models with shared type dependencies."""

from datetime import datetime, timezone
from typing import List, Optional
from uuid import UUID, uuid4

from sqlalchemy import UniqueConstraint
from sqlmodel import Field, Relationship, SQLModel, String

from lightly_purple.server.models.annotation_label import AnnotationLabelBase
from lightly_purple.server.models.annotation_task import (
    AnnotationTask,  # noqa: F401 needed for SQL model generation
    AnnotationType,  # noqa: F401 needed for SQL model generation
)
from lightly_purple.server.models.bounding_box_annotation import (
    BoundingBoxAnnotationBase,
)
from lightly_purple.server.models.dataset import DatasetBase
from lightly_purple.server.models.embedding_model import EmbeddingModelBase
from lightly_purple.server.models.sample import (
    SampleBase,
    SampleViewForAnnotation,
)
from lightly_purple.server.models.sample_embedding import SampleEmbeddingBase
from lightly_purple.server.models.settings import SettingBase
from lightly_purple.server.models.tag import TagBase, TagKind, TagView


class AnnotationTagLink(SQLModel, table=True):
    """AnnotationTagLink links Annotation and Tag Many-to-Many."""

    annotation_id: Optional[UUID] = Field(
        default=None,
        foreign_key="bounding_box_annotations.annotation_id",
        primary_key=True,
    )
    tag_id: Optional[UUID] = Field(
        default=None, foreign_key="tags.tag_id", primary_key=True
    )


class BoundingBoxAnnotation(BoundingBoxAnnotationBase, table=True):
    """This class defines the BoundingBoxAnnotation model."""

    __tablename__ = "bounding_box_annotations"
    annotation_id: UUID = Field(default_factory=uuid4, primary_key=True)
    annotation_label: "AnnotationLabel" = Relationship(
        back_populates="annotations",
        sa_relationship_kwargs={"lazy": "select"},
    )
    sample: Optional["Sample"] = Relationship(
        back_populates="annotations",
        sa_relationship_kwargs={"lazy": "select"},
    )

    """The tag ids associated with the sample."""
    tags: List["Tag"] = Relationship(
        back_populates="annotations", link_model=AnnotationTagLink
    )


class AnnotationLabel(AnnotationLabelBase, table=True):
    """This class defines the AnnotationLabel model."""

    __tablename__ = "annotation_labels"
    annotation_label_id: UUID = Field(default_factory=uuid4, primary_key=True)
    annotations: List["BoundingBoxAnnotation"] = Relationship(
        back_populates="annotation_label",
    )


class SampleTagLink(SQLModel, table=True):
    """SampleTagLink links Sample and Tag Many-to-Many."""

    sample_id: Optional[UUID] = Field(
        default=None, foreign_key="samples.sample_id", primary_key=True
    )
    tag_id: Optional[UUID] = Field(
        default=None, foreign_key="tags.tag_id", primary_key=True
    )


class Sample(SampleBase, table=True):
    """This class defines the Sample model."""

    __tablename__ = "samples"
    sample_id: UUID = Field(default_factory=uuid4, primary_key=True)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )
    annotations: List["BoundingBoxAnnotation"] = Relationship(
        back_populates="sample",
    )

    """The tag ids associated with the sample."""
    tags: List["Tag"] = Relationship(
        back_populates="samples", link_model=SampleTagLink
    )
    embeddings: List["SampleEmbedding"] = Relationship(back_populates="sample")


class Tag(TagBase, table=True):
    """This class defines the Tag model."""

    __tablename__ = "tags"
    # ensure there can only be one tag named "purple" per dataset
    __table_args__ = (
        UniqueConstraint(
            "dataset_id", "kind", "name", name="unique_name_constraint"
        ),
    )
    tag_id: UUID = Field(default_factory=uuid4, primary_key=True)
    dataset_id: UUID
    kind: TagKind = Field(sa_type=String)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )

    """The sample ids associated with the tag."""
    samples: List["Sample"] = Relationship(
        back_populates="tags",
        link_model=SampleTagLink,
    )
    """The annotation ids associated with the tag."""
    annotations: List["BoundingBoxAnnotation"] = Relationship(
        back_populates="tags",
        link_model=AnnotationTagLink,
    )


class Dataset(DatasetBase, table=True):
    """This class defines the Dataset model."""

    __tablename__ = "datasets"
    dataset_id: UUID = Field(default_factory=uuid4, primary_key=True)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )


class AnnotationView(SQLModel):
    """Annotation view model."""

    x: float
    y: float
    width: float
    height: float
    sample_id: UUID
    dataset_id: UUID
    annotation_id: UUID
    annotation_label: "AnnotationLabelView"
    annotation_task_id: UUID
    confidence: Optional[float] = None
    segmentation__binary_mask__rle_row_wise: Optional[List[int]]


class AnnotationViewForAnnotationLabel(SQLModel):
    """Annotation view model for annotation."""

    annotation_id: UUID
    x: float
    y: float
    width: float
    height: float
    sample: "SampleViewForAnnotation"
    annotation_label: "AnnotationLabelView"
    annotation_task_id: UUID
    confidence: Optional[float] = None
    segmentation__binary_mask__rle_row_wise: Optional[List[int]]


class AnnotationLabelView(AnnotationLabelBase):
    """AnnotationLabel class when retrieving."""

    annotation_label_id: UUID


class SampleView(SQLModel):
    """Sample class when retrieving."""

    """The name of the image file."""
    file_name: str
    file_path_abs: str
    sample_id: UUID
    dataset_id: UUID
    annotations: List["AnnotationView"] = Field([])
    tags: List["TagView"] = Field([])
    width: int
    height: int


class EmbeddingModel(EmbeddingModelBase, table=True):
    """This class defines the EmbeddingModel model."""

    __tablename__ = "embedding_models"
    embedding_model_id: UUID = Field(default_factory=uuid4, primary_key=True)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )


class SampleEmbedding(SampleEmbeddingBase, table=True):
    """This class defines the SampleEmbedding model."""

    __tablename__ = "sample_embeddings"
    sample: "Sample" = Relationship(back_populates="embeddings")


class Setting(SettingBase, table=True):
    """This class defines the Setting model."""

    __tablename__ = "settings"
    setting_id: UUID = Field(default_factory=uuid4, primary_key=True)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )


class SettingView(SettingBase):
    """Setting class when retrieving."""

    setting_id: UUID



"""RandomForest classifier implementations."""

from __future__ import annotations

import pickle
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path

import numpy as np
import sklearn  # type: ignore[import-untyped]
from sklearn.ensemble import (  # type: ignore[import-untyped]
    RandomForestClassifier,
)
from sklearn.tree import (  # type: ignore[import-untyped]
    DecisionTreeClassifier,
)
from typing_extensions import assert_never

from .classifier import AnnotatedEmbedding, ExportType, FewShotClassifier

# The version of the file format used for exporting and importing classifiers.
# This is used to ensure compatibility between different versions of the code.
# If the format changes, this version should be incremented.
FILE_FORMAT_VERSION = "1.0.0"


@dataclass
class ModelExportMetadata:
    """Metadata for exporting a model for traceability and reproducibility."""

    file_format_version: str
    model_type: str
    created_at: str
    class_names: list[str]
    num_input_features: int
    num_estimators: int
    embedding_model_hash: str
    embedding_model_name: str
    sklearn_version: str


@dataclass
class InnerNode:
    """Inner node of a decision tree.

    Defaults are used for tree construction.
    """

    feature_index: int = 0
    threshold: float = 0.0
    left_child: int = 0
    right_child: int = 0


@dataclass
class LeafNode:
    """Leaf node of a decision tree."""

    class_probabilities: list[float]


@dataclass
class ExportedTree:
    """Exported tree structure."""

    inner_nodes: list[InnerNode]
    leaf_nodes: list[LeafNode]


@dataclass
class RandomForestExport:
    """Datastructure for exporting the RandomForest model."""

    metadata: ModelExportMetadata
    trees: list[ExportedTree]


class RandomForest(FewShotClassifier):
    """RandomForest classifier."""

    def __init__(
        self,
        classes: list[str],
        embedding_model_name: str,
        embedding_model_hash: str,
    ) -> None:
        """Initialize the RandomForestClassifier with predefined classes.

        Args:
            classes: Ordered list of class labels that will be used for training
                and predictions. The order of this list determines the order of
                probability values in predictions.
            embedding_model_name: Name of the model used for creating the
                embeddings.
            embedding_model_hash: Hash of the model used for creating the
                embeddings.
            Note: embedding_model_name and embedding_model_hash are used for
            traceability in the exported model metadata.

        Raises:
            ValueError: If classes list is empty.
        """
        if not classes:
            raise ValueError("Class list cannot be empty.")

        # Fix the random seed for reproducibility.
        self._model = RandomForestClassifier(
            class_weight="balanced", random_state=42
        )
        self._classes = classes
        self._class_to_index = {label: idx for idx, label in enumerate(classes)}
        self._embedding_model_name = embedding_model_name
        self._embedding_model_hash = embedding_model_hash or ""

    def train(self, annotated_embeddings: list[AnnotatedEmbedding]) -> None:
        """Trains a classifier using the provided input.

        Args:
            annotated_embeddings: A list of annotated embeddings to train the
            classifier.

        Raises:
            ValueError: If annotated_embeddings is empty or contains invalid
            classes.
        """
        if not annotated_embeddings:
            raise ValueError("annotated_embeddings cannot be empty.")

        # Extract embeddings and labels.
        embeddings = [ae.embedding for ae in annotated_embeddings]
        labels = [ae.annotation for ae in annotated_embeddings]
        # Validate that all labels are in predefined classes.
        invalid_labels = set(labels) - set(self._classes)
        if invalid_labels:
            raise ValueError(
                f"Found labels not in predefined classes: {invalid_labels}"
            )

        # Convert to NumPy arrays.
        embeddings_np = np.array(embeddings)
        labels_encoded = [self._class_to_index[label] for label in labels]

        # Train the RandomForestClassifier.
        self._model.fit(embeddings_np, labels_encoded)

    def predict(self, embeddings: list[list[float]]) -> list[list[float]]:
        """Predicts the classification scores for a list of embeddings.

        Args:
            embeddings: A list of embeddings, where each embedding is a list of
            floats.

        Returns:
            A list of lists, where each inner list represents the probability
            distribution over classes for the corresponding input embedding.
            Each value in the inner list corresponds to the likelihood of the
            embedding belonging to a specific class.
            If embeddings is empty, returns an empty list.
        """
        if len(embeddings) == 0:
            return []

        # Convert embeddings to a NumPy array.
        embeddings_np = np.array(embeddings)

        # Get the classes that the model was trained on.
        trained_classes: list[int] = self._model.classes_

        # Initialize full-size probability array.
        full_probabilities = []

        # Get raw probabilities from model.
        raw_probabilities = self._model.predict_proba(embeddings_np)

        for raw_probs in raw_probabilities:
            # Initialize zeros for all possible classes.
            full_probs = [0.0 for _ in range(len(self._classes))]
            # Map probabilities to their correct positions.
            for trained_class, prob in zip(trained_classes, raw_probs):
                full_probs[trained_class] = prob
            full_probabilities.append(full_probs)
        return full_probabilities

    def export(
        self, export_path: Path, export_type: ExportType = "sklearn"
    ) -> None:
        """Exports the classifier to a specified file.

        Args:
            export_path: The full file path where the export will be saved.
            export_type: The type of export. Options are:
                "sklearn": Exports the RandomForestClassifier instance.
                "lightly": Exports the model in raw format with metadata
                and tree details.
        """
        metadata = ModelExportMetadata(
            file_format_version=FILE_FORMAT_VERSION,
            model_type="RandomForest",
            created_at=str(datetime.now(timezone.utc).isoformat()),
            class_names=self._classes,
            num_input_features=self._model.n_features_in_,
            num_estimators=len(self._model.estimators_),
            embedding_model_hash=self._embedding_model_hash,
            embedding_model_name=self._embedding_model_name,
            sklearn_version=sklearn.__version__,
        )

        if export_type == "sklearn":
            # Combine the model and metadata into a single dictionary
            export_data = {
                "model": self._model,
                "metadata": metadata,
            }
            # Save to the specified file path.
            # Ensure parent dirs exist.
            export_path.parent.mkdir(parents=True, exist_ok=True)
            with open(export_path, "wb") as f:
                pickle.dump(export_data, f)

        elif export_type == "lightly":
            export_data_raw = _export_random_forest_model(
                model=self._model,
                metadata=metadata,
                all_classes=self._classes,
            )
            # Save to the specified file path.
            # Ensure parent dirs exist.
            export_path.parent.mkdir(parents=True, exist_ok=True)
            with open(export_path, "wb") as f:
                pickle.dump(export_data_raw, f)
        else:
            assert_never(export_type)


def load_random_forest_classifier(classifier_path: Path) -> RandomForest:
    """Loads a RandomForest classifier from a file.

    Args:
        classifier_path: The path to the exported classifier file.

    Returns:
        A fully initialized RandomForest classifier instance.

    Raises:
        FileNotFoundError: If the classifier_path does not exist.
        ValueError: If the file is not a valid 'sklearn' pickled export
                    or if the version/format mismatches.
    """
    if not classifier_path.exists():
        raise FileNotFoundError(f"The file {classifier_path} does not exist.")

    with open(classifier_path, "rb") as f:
        export_data = pickle.load(f)

    model = export_data.get("model")
    metadata: ModelExportMetadata = export_data.get("metadata")

    if model is None or metadata is None:
        raise ValueError(
            "The loaded file does not contain a valid model or metadata."
        )

    if metadata.file_format_version != FILE_FORMAT_VERSION:
        raise ValueError(
            f"File format version mismatch. Expected '{FILE_FORMAT_VERSION}', "
            f"got '{metadata.file_format_version}'."
        )
    instance = RandomForest(
        classes=metadata.class_names,
        embedding_model_name=metadata.embedding_model_name,
        embedding_model_hash=metadata.embedding_model_hash,
    )
    # Set the model.
    instance._model = model  # noqa: SLF001
    return instance


def _export_random_forest_model(
    model: RandomForestClassifier,
    metadata: ModelExportMetadata,
    all_classes: list[str],
) -> RandomForestExport:
    """Converts a sk-learn RandomForestClassifier to RandomForestExport format.

    Args:
        model: The trained random forest model to export.
        metadata: Metadata describing the dataset and training setup.
        all_classes: Full list of all class labels.

    Returns:
        RandomForestExport: The serialized export object containing all trees
            and metadata.
    """
    trained_classes: list[int] = model.classes_
    trees = [
        _export_single_tree(tree, trained_classes, all_classes)
        for tree in model.estimators_
    ]
    return RandomForestExport(metadata=metadata, trees=trees)


def load_lightly_random_forest(path: Path) -> RandomForestExport:
    """Loads a Lightly exported RandomForest model from a file.

    Args:
        path: The path to the exported classifier file.

    Returns:
        A RandomForestExport instance.

    Raises:
        ValueError: If the file is not a valid RandomForestExport or
                if the version/format mismatches.
    """
    with open(path, "rb") as f:
        data = pickle.load(f)
    if not isinstance(data, RandomForestExport):
        raise ValueError("Loaded object is not a RandomForestExport instance.")

    if data.metadata.file_format_version != FILE_FORMAT_VERSION:
        raise ValueError(
            f"File format version mismatch. Expected '{FILE_FORMAT_VERSION}', "
            f"got '{data.metadata.file_format_version}'."
        )
    return data


def predict_with_lightly_random_forest(
    model: RandomForestExport, embeddings: list[list[float]]
) -> list[list[float]]:
    """Predicts the classification scores for a list of embeddings.

    Args:
        model: A RandomForestExport instance containing the model and metadata.
        embeddings: A list of embeddings.

    Returns:
        A list of lists, where each inner list represents the probability
            distribution over classes for the corresponding input embedding.

    Raises:
        ValueError: If the provided embeddings have different size than
            expected.
    """
    expected_dim = model.metadata.num_input_features
    all_probs: list[list[float]] = []

    for embedding in embeddings:
        if len(embedding) != expected_dim:
            raise ValueError(
                f"Embedding has wrong dimensionality: expected {expected_dim},"
                f"got {len(embedding)}"
            )

        tree_probs: list[list[float]] = [
            _predict_tree_probs(tree, embedding) for tree in model.trees
        ]

        mean_probs = np.mean(tree_probs, axis=0).tolist()
        all_probs.append(mean_probs)

    return all_probs


def _export_single_tree(
    tree: DecisionTreeClassifier,
    trained_classes: list[int],
    all_classes: list[str],
) -> ExportedTree:
    """Converts a single sk-learn tree into a serializable ExportedTree format.

    Args:
        tree: The decision tree to convert.
        trained_classes: Indices of the classes the tree was trained on.
        all_classes: Full list of all class labels.

    Returns:
        ExportedTree: A representation of the tree with explicit node and leaf
                    structures, compatible with the Lightly format.
    """
    tree_structure = tree.tree_
    inner_nodes: list[InnerNode] = []
    leaf_nodes: list[LeafNode] = []
    node_map = {}  # Maps node_id to (mapped_index, is_leaf)

    for node_id in range(tree_structure.node_count):
        is_leaf = (
            tree_structure.children_left[node_id]
            == tree_structure.children_right[node_id]
        )
        if is_leaf:
            leaf_idx = len(leaf_nodes)
            # value[node_id] is a 2D array of shape [1, n_classes].
            # [0] is used to extract the inner array and
            # convert it to a 1D array of class counts.
            class_weights = tree_structure.value[node_id][0]
            total = sum(class_weights)
            probs = (
                (class_weights / total).tolist()
                if total > 0
                else [0.0] * len(class_weights)
            )

            # Order probabilities according to the initial classes.
            # Initialize zeros for all possible classes.
            full_probs = [0.0 for _ in range(len(all_classes))]
            # Map probabilities to their correct positions.
            for trained_class, prob in zip(trained_classes, probs):
                full_probs[trained_class] = prob

            leaf_nodes.append(LeafNode(class_probabilities=full_probs))
            node_map[node_id] = (-leaf_idx - 1, True)
        else:
            inner_idx = len(inner_nodes)
            node_map[node_id] = (inner_idx, False)
            # Reserve a spot for the inner node.
            inner_nodes.append(InnerNode())

    # Now populate inner_nodes using mapped indices.
    for node_id in range(tree_structure.node_count):
        mapped_idx, is_leaf = node_map[node_id]
        if is_leaf:
            continue

        left_id = tree_structure.children_left[node_id]
        right_id = tree_structure.children_right[node_id]
        left_mapped = node_map[left_id][0]
        right_mapped = node_map[right_id][0]

        inner_nodes[mapped_idx] = InnerNode(
            feature_index=int(tree_structure.feature[node_id]),
            threshold=float(tree_structure.threshold[node_id]),
            left_child=left_mapped,
            right_child=right_mapped,
        )

    return ExportedTree(inner_nodes=inner_nodes, leaf_nodes=leaf_nodes)


def _predict_tree_probs(
    tree: ExportedTree, embedding: list[float]
) -> list[float]:
    """Predicts class probabilities for an embedding using a single tree.

    Args:
        tree: A ExportedTree instance used to determine the probability.
        embedding: A single embedding.

    """
    if not tree.inner_nodes:
        return tree.leaf_nodes[0].class_probabilities

    node_idx = 0  # Start at root
    while node_idx >= 0:
        node = tree.inner_nodes[node_idx]
        if embedding[node.feature_index] <= node.threshold:
            node_idx = node.left_child
        else:
            node_idx = node.right_child

    leaf_idx = -node_idx - 1
    leaf = tree.leaf_nodes[leaf_idx]
    return leaf.class_probabilities



"""This module defines the AnnotationTaskResolver."""

from typing import List, Optional, Sequence
from uuid import UUID

from sqlmodel import Session, select

from lightly_purple.server.models.annotation_task import AnnotationTask


class AnnotationTaskResolver:
    """Resolver class for AnnotationTask models."""

    def __init__(self, session: Session) -> None:
        """Initialize the resolver with a database session."""
        self.session = session

    def create(self, annotation_task: AnnotationTask) -> AnnotationTask:
        """Create a new annotation task."""
        self.session.add(annotation_task)
        self.session.commit()
        self.session.refresh(annotation_task)
        return annotation_task

    def get_by_id(self, annotation_task_id: UUID) -> Optional[AnnotationTask]:
        """Get an annotation task by ID."""
        statement = select(AnnotationTask).where(
            AnnotationTask.annotation_task_id == annotation_task_id
        )
        return self.session.exec(statement).first()

    def get_all(self) -> List[AnnotationTask]:
        """Get all annotation tasks."""
        statement = select(AnnotationTask)
        results: Sequence[AnnotationTask] = self.session.exec(statement).all()
        return list(results)



"""This module contains the API routes for user settings."""

from fastapi import APIRouter, Depends
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import SettingView
from lightly_purple.server.resolvers.settings import SettingsResolver

settings_router = APIRouter(tags=["settings"])

SessionDep = Annotated[Session, Depends(get_session)]


def get_settings_resolver(session: SessionDep) -> SettingsResolver:
    """Create an instance of the SettingsResolver."""
    return SettingsResolver(session)


@settings_router.get("/settings")
def get_settings(
    handler: Annotated[SettingsResolver, Depends(get_settings_resolver)],
) -> SettingView:
    """Get the current settings.

    Args:
        handler: Settings resolver instance for database operations.

    Returns:
        The current settings.
    """
    return handler.get_settings()


@settings_router.post("/settings")
def set_settings(
    settings: SettingView,
    handler: Annotated[SettingsResolver, Depends(get_settings_resolver)],
) -> SettingView:
    """Update user settings.

    Args:
        settings: New settings to apply.
        handler: Settings resolver instance for database operations.

    Returns:
        Updated settings.
    """
    return handler.set_settings(settings)



"""MobileCLIP embedding generator."""

from __future__ import annotations

import tempfile
from pathlib import Path
from uuid import UUID

import mobileclip  # type: ignore[import-untyped]
import torch
from PIL import Image

from lightly_purple.server.models.embedding_model import EmbeddingModelInput

from . import file_utils
from .embedding_generator import EmbeddingGenerator

MODEL_NAME = "mobileclip_s0"
MOBILECLIP_DOWNLOAD_URL = f"https://docs-assets.developer.apple.com/ml-research/datasets/mobileclip/{MODEL_NAME}.pt"


class MobileCLIPEmbeddingGenerator(EmbeddingGenerator):
    """MobileCLIP embedding model."""

    def __init__(self) -> None:
        """Initialize the MobileCLIP embedding model.

        This method loads the MobileCLIP model and its tokenizer. The model
        checkpoint is downloaded and cached locally for future use.
        """
        model_path = _get_cached_mobileclip_checkpoint()
        self._model, _, self._preprocess = (
            mobileclip.create_model_and_transforms(
                model_name=MODEL_NAME, pretrained=model_path
            )
        )
        self._tokenizer = mobileclip.get_tokenizer(model_name=MODEL_NAME)
        self._model_hash = file_utils.get_file_xxhash(model_path)

    def get_embedding_model_input(
        self, dataset_id: UUID
    ) -> EmbeddingModelInput:
        """Generate an EmbeddingModelInput instance.

        Args:
            dataset_id: The ID of the dataset.

        Returns:
            An EmbeddingModelInput instance with the model details.
        """
        return EmbeddingModelInput(
            name=MODEL_NAME,
            embedding_model_hash=self._model_hash,
            embedding_dimension=512,
            dataset_id=dataset_id,
        )

    def embed_text(self, text: str) -> list[float]:
        """Embed a text with MobileCLIP.

        Args:
            text: The text to embed.

        Returns:
            A list of floats representing the generated embedding.
        """
        tokenized = self._tokenizer([text])
        with torch.no_grad():
            embedding = self._model.encode_text(tokenized)[0]
            # Convert embedding to list of floats.
            embedding_list: list[float] = (
                embedding.cpu().numpy().flatten().tolist()
            )
        return embedding_list

    def embed_images(self, filepaths: list[Path]) -> list[list[float]]:
        """Embed images with MobileCLIP.

        Args:
            filepaths: A list of file paths to the images to embed.

        Returns:
            A list of lists of floats representing the generated embeddings
            in the same order as the input file paths.
        """
        images = [
            self._preprocess(Image.open(filepath).convert("RGB"))
            for filepath in filepaths
        ]
        images_tensor = torch.stack(images)
        with torch.no_grad():
            embeddings = self._model.encode_image(images_tensor)
            # Convert embedding to list of floats.
            embeddings_list: list[list[float]] = (
                embeddings.cpu().numpy().tolist()
            )
        return embeddings_list


def _get_cached_mobileclip_checkpoint() -> Path:
    file_path = Path(tempfile.gettempdir()) / "mobileclip_s0.pt"
    file_utils.download_file_if_does_not_exist(
        url=MOBILECLIP_DOWNLOAD_URL,
        local_filename=file_path,
    )
    return file_path



"""Handler for database operations related to annotations."""

from __future__ import annotations

from uuid import UUID

from pydantic import BaseModel, Field
from sqlmodel import Session, func, select

from lightly_purple.server.models import (
    AnnotationLabel,
    BoundingBoxAnnotation,
    Sample,
    Tag,
)
from lightly_purple.server.models.bounding_box_annotation import (
    BoundingBoxAnnotationInput,
)
from lightly_purple.server.routes.api.validators import Paginated


class AnnotationsFilterParams(BaseModel):
    """Encapsulates filter parameters for querying annotations."""

    dataset_ids: list[UUID] | None = Field(
        default=None, description="List of dataset UUIDs"
    )
    annotation_label_ids: list[UUID] | None = Field(
        default=None, description="List of annotation label UUIDs"
    )
    annotation_tag_ids: list[UUID] | None = Field(
        default=None, description="List of tag UUIDs"
    )
    sample_tag_ids: list[UUID] | None = Field(
        default=None,
        description="List of sample tag UUIDs to filter annotations by",
    )
    annotation_task_ids: list[UUID] | None = Field(
        default=None, description="List of annotation task UUIDs"
    )


class BoundingBoxAnnotationResolver:
    """Resolver for the BoundingBoxAnnotation model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(
        self, annotation: BoundingBoxAnnotationInput
    ) -> BoundingBoxAnnotation:
        """Create a new annotation in the database."""
        db_annotation = BoundingBoxAnnotation.model_validate(annotation)
        self.session.add(db_annotation)
        self.session.commit()
        self.session.refresh(db_annotation)
        return db_annotation

    def create_many(
        self, annotations: list[BoundingBoxAnnotationInput]
    ) -> None:
        """Create many annotations in a single commit. No return values."""
        db_annotations = [
            BoundingBoxAnnotation.model_validate(a) for a in annotations
        ]
        self.session.bulk_save_objects(db_annotations)
        self.session.commit()

    def get_all(  # noqa: C901, PLR0912
        self,
        pagination: dict | Paginated | None = None,
        filters: AnnotationsFilterParams | None = None,
    ) -> list[BoundingBoxAnnotation]:
        """Retrieve all annotations from the database."""
        # Determine pagination parameters
        if pagination is None:
            offset = None
            limit = None
        elif isinstance(pagination, dict):
            pag = Paginated(**pagination)
            offset = pag.offset
            limit = pag.limit
        else:
            offset = pagination.offset
            limit = pagination.limit

        if filters is None:
            filters = AnnotationsFilterParams()
        elif isinstance(filters, dict):
            filters = AnnotationsFilterParams(**filters)

        query = select(BoundingBoxAnnotation)

        # Apply filters if provided
        if filters:
            if filters.dataset_ids:
                query = query.where(
                    BoundingBoxAnnotation.dataset_id.in_(filters.dataset_ids)
                )

            if filters.annotation_task_ids:
                query = query.where(
                    BoundingBoxAnnotation.annotation_task_id.in_(
                        filters.annotation_task_ids
                    )
                )

            if filters.annotation_label_ids:
                query = query.where(
                    BoundingBoxAnnotation.annotation_label_id.in_(
                        filters.annotation_label_ids
                    )
                )

            # Filter by annotation tags
            if filters.annotation_tag_ids:
                query = (
                    query.join(BoundingBoxAnnotation.tags)
                    .where(
                        BoundingBoxAnnotation.tags.any(
                            Tag.tag_id.in_(filters.annotation_tag_ids)
                        )
                    )
                    .distinct()
                )

            # Filter by sample tags
            if filters.sample_tag_ids:
                query = (
                    query.join(
                        Sample,
                        BoundingBoxAnnotation.sample_id == Sample.sample_id,
                    )
                    .join(Sample.tags)
                    .where(
                        Sample.tags.any(Tag.tag_id.in_(filters.sample_tag_ids))
                    )
                    .distinct()
                )

        # Apply pagination if specified
        if offset is not None:
            query = query.offset(offset)
        if limit is not None:
            query = query.limit(limit)

        results = self.session.exec(query).all()
        return results or []

    def get_by_id(self, annotation_id: UUID) -> BoundingBoxAnnotation | None:
        """Retrieve a single annotation by ID."""
        return self.session.exec(
            select(BoundingBoxAnnotation).where(
                BoundingBoxAnnotation.annotation_id == annotation_id
            )
        ).one_or_none()

    def update(
        self, annotation_id: UUID, annotation_data: BoundingBoxAnnotationInput
    ) -> BoundingBoxAnnotation | None:
        """Update an existing annotation."""
        annotation = self.get_by_id(annotation_id)
        if not annotation:
            return None

        annotation.x = annotation_data.x
        annotation.y = annotation_data.y
        annotation.width = annotation_data.width
        annotation.height = annotation_data.height

        # Update new fields if they're provided
        if (
            hasattr(annotation_data, "confidence")
            and annotation_data.confidence is not None
        ):
            annotation.confidence = annotation_data.confidence

        if (
            hasattr(annotation_data, "annotation_task_id")
            and annotation_data.annotation_task_id is not None
        ):
            annotation.annotation_task_id = annotation_data.annotation_task_id

        if (
            hasattr(annotation_data, "segmentation__binary_mask__rle_row_wise")
            and annotation_data.segmentation__binary_mask__rle_row_wise
            is not None
        ):
            annotation.segmentation__binary_mask__rle_row_wise = (
                annotation_data.segmentation__binary_mask__rle_row_wise
            )

        self.session.commit()
        self.session.refresh(annotation)
        return annotation

    def delete(self, annotation_id: UUID) -> bool:
        """Delete an annotation."""
        annotation = self.get_by_id(annotation_id)
        if not annotation:
            return False

        self.session.delete(annotation)
        self.session.commit()
        return True

    def count_annotations_by_dataset(  # noqa: PLR0913 // FIXME: refactor to use proper pydantic
        self,
        dataset_id: UUID,
        filtered_labels: list[str] | None = None,
        min_width: int | None = None,
        max_width: int | None = None,
        min_height: int | None = None,
        max_height: int | None = None,
        tag_ids: list[UUID] | None = None,
    ) -> list[tuple[str, int, int]]:
        """Count annotations for a specific dataset.

        Annotations for a specific dataset are grouped by annotation
        label name and counted for total and filtered.
        """
        # Query for total counts (unfiltered)
        total_counts_query = (
            select(
                AnnotationLabel.annotation_label_name,
                func.count(BoundingBoxAnnotation.annotation_id).label(
                    "total_count"
                ),
            )
            .join(
                BoundingBoxAnnotation,
                BoundingBoxAnnotation.annotation_label_id
                == AnnotationLabel.annotation_label_id,
            )
            .join(Sample, Sample.sample_id == BoundingBoxAnnotation.sample_id)
            .where(Sample.dataset_id == dataset_id)
            .group_by(AnnotationLabel.annotation_label_name)
            .order_by(AnnotationLabel.annotation_label_name.asc())
        )

        total_counts = {
            row[0]: row[1]
            for row in self.session.exec(total_counts_query).all()
        }

        # Build filtered query for current counts
        filtered_query = (
            select(
                AnnotationLabel.annotation_label_name,
                func.count(BoundingBoxAnnotation.annotation_id).label(
                    "current_count"
                ),
            )
            .join(
                BoundingBoxAnnotation,
                BoundingBoxAnnotation.annotation_label_id
                == AnnotationLabel.annotation_label_id,
            )
            .join(Sample, Sample.sample_id == BoundingBoxAnnotation.sample_id)
            .where(Sample.dataset_id == dataset_id)
        )

        # Add dimension filters
        if min_width is not None:
            filtered_query = filtered_query.where(Sample.width >= min_width)
        if max_width is not None:
            filtered_query = filtered_query.where(Sample.width <= max_width)
        if min_height is not None:
            filtered_query = filtered_query.where(Sample.height >= min_height)
        if max_height is not None:
            filtered_query = filtered_query.where(Sample.height <= max_height)

        # Add label filter if specified
        if filtered_labels:
            filtered_query = filtered_query.where(
                Sample.sample_id.in_(
                    select(Sample.sample_id)
                    .join(
                        BoundingBoxAnnotation,
                        Sample.sample_id == BoundingBoxAnnotation.sample_id,
                    )
                    .join(
                        AnnotationLabel,
                        BoundingBoxAnnotation.annotation_label_id
                        == AnnotationLabel.annotation_label_id,
                    )
                    .where(
                        AnnotationLabel.annotation_label_name.in_(
                            filtered_labels
                        )
                    )
                )
            )

        # filter by tag_ids
        if tag_ids:
            filtered_query = (
                filtered_query.join(BoundingBoxAnnotation.tags)
                .where(BoundingBoxAnnotation.tags.any(Tag.tag_id.in_(tag_ids)))
                .distinct()
            )

        # Group by label name and sort
        filtered_query = filtered_query.group_by(
            AnnotationLabel.annotation_label_name
        ).order_by(AnnotationLabel.annotation_label_name.asc())

        _rows = self.session.exec(filtered_query).all()

        current_counts = {row[0]: row[1] for row in _rows}

        return [
            (label, current_counts.get(label, 0), total_count)
            for label, total_count in total_counts.items()
        ]



"""This module contains the API routes for managing text embedding."""

from __future__ import annotations

from typing import List
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query
from typing_extensions import Annotated

from lightly_purple.dataset.embedding_manager import (
    EmbeddingManager,
    EmbeddingManagerProvider,
    TextEmbedQuery,
)
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_INTERNAL_SERVER_ERROR,
)

text_embedding_router = APIRouter()
# Define a type alias for the EmbeddingManager dependency
EmbeddingManagerDep = Annotated[
    EmbeddingManager,
    Depends(lambda: EmbeddingManagerProvider.get_embedding_manager(None)),
]


@text_embedding_router.get(
    "/text_embedding/embed_text", response_model=List[float]
)
def embed_text(
    embedding_manager: EmbeddingManagerDep,
    query_text: str = Query(..., description="The text to embed."),
    embedding_model_id: Annotated[
        UUID | None,
        Query(..., description="The ID of the embedding model to use."),
    ] = None,
) -> list[float]:
    """Retrieve embeddings for the input text."""
    try:
        text_embeddings = embedding_manager.embed_text(
            TextEmbedQuery(query_text, embedding_model_id)
        )
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_STATUS_INTERNAL_SERVER_ERROR,
            detail=f"{exc}",
        ) from None

    return text_embeddings



"""Global list fo the active features."""

# TODO(Kondrat 04/25): Pass the feature flag to app
# https://linear.app/lightly/issue/LIG-6708/introduce-apifeatures-endpoint
from typing import List

purple_active_features: List[str] = []



"""This module contains the API routes for active features."""

from __future__ import annotations

from fastapi import APIRouter

from lightly_purple.server.features import purple_active_features

__all__ = ["features_router", "purple_active_features"]

features_router = APIRouter()


@features_router.get("/features")
def get_features() -> list[str]:
    """Get the list of active features in the Purple app."""
    return purple_active_features



"""Module provides functions to initialize and manage the DuckDB."""

import logging
import os
from contextlib import contextmanager
from typing import Generator

from sqlmodel import Session, SQLModel, create_engine

import lightly_purple.server.models  # noqa: F401, required for SQLModel to work properly


class MockDatabaseManager:
    """Mock version of DatabaseManager."""

    def __init__(self) -> None:
        """Create a new instance of the MockDatabaseManager."""
        self.engine = create_engine("duckdb:///:memory:")
        self._session_instance = Session(self.engine)
        # Initialize tables
        SQLModel.metadata.create_all(self.engine)

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        """Return the database session."""
        try:
            yield self._session_instance
            if not self._session_instance.in_transaction():
                self._session_instance.commit()
        except Exception:
            self._session_instance.rollback()
            raise
        finally:
            self._session_instance.close()


class DatabaseManager:
    """Manages database connections and ensures proper resource handling."""

    _instance = None

    @staticmethod
    def database_exists(db_file: str) -> bool:
        """Check if database file exists.

        Args:
            db_file: Path to the database file

        Returns:
            True if database file exists, False otherwise
        """
        return os.path.exists(db_file)

    @staticmethod
    def cleanup_database(db_file: str) -> None:
        """Delete database file if it exists.

        Args:
            db_file: Path to the database file to delete
        """
        if DatabaseManager.database_exists(db_file):
            os.remove(db_file)
            logging.info(f"Deleted existing database: {db_file}")

    def __new__(
        cls, db_file: str = "purple.db", cleanup_existing: bool = False
    ):
        """Create a new instance of the DatabaseManager.

        Args:
            db_file: Path to the database file
            cleanup_existing: If True, deletes existing database
            before creating a new one

        Returns:
            DatabaseManager instance
        """
        if cleanup_existing:
            cls.cleanup_database(db_file)

        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # File-based DuckDB
            cls._instance.engine = create_engine(f"duckdb:///{db_file}")
            # Initialize tables
            SQLModel.metadata.create_all(cls._instance.engine)
        return cls._instance

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        """Create a new database session."""
        session = Session(self.engine)
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()


# Global instance
db_manager = DatabaseManager(cleanup_existing=True)


# For FastAPI dependency injection
def get_session():
    """Yield a new session for database operations."""
    with db_manager.session() as session:
        yield session



"""This module contains settings model for user preferences."""

from enum import Enum

from sqlmodel import Field, SQLModel


class GridViewSampleRenderingType(str, Enum):
    """Defines how samples are rendered in the grid view."""

    COVER = "cover"
    CONTAIN = "contain"


class SettingBase(SQLModel):
    """Base class for Settings model."""

    grid_view_sample_rendering: GridViewSampleRenderingType = Field(
        default=GridViewSampleRenderingType.CONTAIN,
        description="Controls how samples are rendered in the grid view",
    )



"""Handler for database operations related to embedding models."""

from __future__ import annotations

from uuid import UUID

from sqlmodel import Session, select

from lightly_purple.server.models import (
    EmbeddingModel,
)
from lightly_purple.server.models.embedding_model import (
    EmbeddingModelInput,
)


class EmbeddingModelResolver:
    """Resolver for the EmbeddingModel model."""

    def __init__(self, session: Session):
        """Initialize the EmbeddingModelResolver."""
        self.session = session

    def create(self, embedding_model: EmbeddingModelInput) -> EmbeddingModel:
        """Create a new EmbeddingModel in the database."""
        db_embedding_model = EmbeddingModel.model_validate(embedding_model)
        self.session.add(db_embedding_model)
        self.session.commit()
        self.session.refresh(db_embedding_model)
        return db_embedding_model

    def get_all_by_dataset_id(self, dataset_id: UUID) -> list[EmbeddingModel]:
        """Retrieve all embedding models."""
        embedding_models = self.session.exec(
            select(EmbeddingModel).where(
                EmbeddingModel.dataset_id == dataset_id
            )
        ).all()
        return list(embedding_models)

    def get_by_id(self, embedding_model_id: UUID) -> EmbeddingModel | None:
        """Retrieve a single embedding model by ID."""
        return self.session.exec(
            select(EmbeddingModel).where(
                EmbeddingModel.embedding_model_id == embedding_model_id
            )
        ).one_or_none()

    def delete(self, embedding_model_id: UUID) -> bool:
        """Delete an embedding model."""
        embedding_model = self.get_by_id(embedding_model_id)
        if not embedding_model:
            return False

        self.session.delete(embedding_model)
        self.session.commit()
        return True



"""FewShotClassifier Protocol."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Protocol, runtime_checkable

ExportType = Literal["sklearn", "lightly"]


@dataclass
class AnnotatedEmbedding:
    """Annotated embedding class.

    Attributes:
        embedding: The embedding vector as a list of floats.
        annotation: The annotation for the embedding.
                    Note: Currently it is an string but it can be a integer or
                    any other type.
    """

    embedding: list[float]
    annotation: str


@runtime_checkable
class FewShotClassifier(Protocol):
    """Protocol defining the interface for classifiers."""

    def train(self, annotated_embeddings: list[AnnotatedEmbedding]) -> None:
        """Trains a classifier using the provided input.

        Args:
            annotated_embeddings: A list of annotated embeddings to train the
            classifier.

        Raises:
            ValueError: If annotated_embeddings is empty or contains invalid
            classes.
        """
        ...

    def predict(self, embeddings: list[list[float]]) -> list[list[float]]:
        """Predicts the classification scores for a list of embeddings.

        Args:
            embeddings: A list of embeddings, where each embedding is a list of
            floats.

        Returns:
            A list of lists, where each inner list represents the probability
            distribution over classes for the corresponding input embedding.
            Each value in the inner list corresponds to the likelihood of the
            embedding belonging to a specific class.
            If embeddings is empty, returns an empty list.
        """
        ...

    def export(self, export_file_path: Path, export_type: ExportType) -> None:
        """Exports the classifier to a specified file.

        This method saves the trained classifier to the given file path so that
        it can be reused or shared.

        Args:
            export_file_path: The path to the file where the classifier should
            be saved.
            export_type: The type of export format to use. This can be either
            "sklearn_instance" or "raw".
        """
        ...
